#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   03/02/2007
#
#-------------------------------------------------------------------------

""" Traits UI 'display only' LED numeric editor.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from wx.gizmos \
    import LEDNumberCtrl, LED_ALIGN_LEFT, LED_ALIGN_CENTER, LED_ALIGN_RIGHT

from traits.api \
    import Enum

from traitsui.wx.editor \
    import Editor

from traitsui.basic_editor_factory \
    import BasicEditorFactory

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# LED alignment styles:
LEDStyles = {
    'left': LED_ALIGN_LEFT,
    'center': LED_ALIGN_CENTER,
    'right': LED_ALIGN_RIGHT,
}

#-------------------------------------------------------------------------
#  '_LEDEditor' class:
#-------------------------------------------------------------------------


class _LEDEditor(Editor):
    """ Traits UI 'display only' LED numeric editor.
    """

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = LEDNumberCtrl(parent, -1)
        self.control.SetAlignment(LEDStyles[self.factory.alignment])
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self.control.SetValue(self.str_value)

#-------------------------------------------------------------------------
#  Create the editor factory object:
#-------------------------------------------------------------------------

# wxPython editor factory for LED editors:


class LEDEditor(BasicEditorFactory):

    # The editor class to be created:
    klass = _LEDEditor

    # The alignment of the numeric text within the control:
    alignment = Enum('right', 'left', 'center')



#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   01/10/2006
#
#------------------------------------------------------------------------------

""" Defines the array editor factory for all traits toolkit backends.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

import numpy

from traits.api import Bool, HasTraits, Int, Float, Instance, false, TraitError

from ..editor import Editor

from ..editor_factory import EditorFactory

# CIRCULAR IMPORT FIXME: Importing from the source rather than traits.ui.api
# to avoid circular imports, as this EditorFactory will be part of
# traits.ui.api as well.
from ..view import View

from ..group import Group

from ..item import Item

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for array editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Width of the individual fields
    width = Int(-80)

    # Is user input set on every keystroke?
    auto_set = Bool(True)

    # Is user input set when the Enter key is pressed?
    enter_set = Bool(False)

#-------------------------------------------------------------------------
#  'ArrayStructure' class:
#-------------------------------------------------------------------------


class ArrayStructure(HasTraits):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Editor that this structure is linked to
    editor = Instance(Editor)

    # The constructed View for the array
    view = Instance(View)

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, editor):
        """ Initializes the object.
        """
        # Save the reference to the editor:
        self.editor = editor

        # Set up the field width for each item:
        width = editor.factory.width

        # Set up the correct style for each filed:
        style = 'simple'
        if editor.readonly:
            style = 'readonly'

        # Get the array we are mirroring:
        object = editor.value

        # Determine the correct trait type to use for each element:
        trait = Float

        if object.dtype.type == 'i':
            trait = Int

        if len(object.shape) == 1:
            self.view = self._one_dim_view(object, style, width, trait)
        elif len(object.shape) == 2:
            self.view = self._two_dim_view(object, style, width, trait)
        else:
            raise TraitError('Only 1D or 2D arrays supported')

    #-------------------------------------------------------------------------
    #  1D view:
    #-------------------------------------------------------------------------

    def _one_dim_view(self, object, style, width, trait):
        content = []
        shape = object.shape
        items = []
        format_func = self.editor.factory.format_func
        format_str = self.editor.factory.format_str
        for i in range(shape[0]):
            name = 'f%d' % i
            self.add_trait(name,
                           trait(object[i],
                                 event='field',
                                 auto_set=self.editor.factory.auto_set,
                                 enter_set=self.editor.factory.enter_set)
                           )
            items.append(Item(name=name,
                              style=style,
                              width=width,
                              format_func=format_func,
                              format_str=format_str,
                              padding=-3))

        group = Group(orientation='horizontal',
                      show_labels=False,
                      *items)
        content.append(group)

        return View(Group(show_labels=False, *content))

    #-------------------------------------------------------------------------
    #  2D view:
    #-------------------------------------------------------------------------

    def _two_dim_view(self, object, style, width, trait):
        content = []
        shape = object.shape
        format_func = self.editor.factory.format_func
        format_str = self.editor.factory.format_str
        for i in range(shape[0]):
            items = []
            for j in range(shape[1]):
                name = 'f%d_%d' % (i, j)
                self.add_trait(name,
                               trait(object[i, j],
                                     event='field',
                                     auto_set=self.editor.factory.auto_set,
                                     enter_set=self.editor.factory.enter_set)
                               )
                items.append(Item(name=name,
                                  style=style,
                                  width=width,
                                  format_func=format_func,
                                  format_str=format_str,
                                  padding=-3))

            group = Group(orientation='horizontal',
                          show_labels=False,
                          *items)
            content.append(group)

        return View(Group(show_labels=False, *content))

    #-------------------------------------------------------------------------
    #  Updates the underlying tuple when any field changes value:
    #-------------------------------------------------------------------------

    def _field_changed(self):
        """ Updates the underlying array when any field changes value.
        """

        if not self.editor._busy:
            # Get the array we are mirroring:
            object = self.editor.value
            shape = object.shape
            value = numpy.zeros(shape, object.dtype)

            # 1D
            if len(shape) == 1:
                for i in range(shape[0]):
                    value[i] = getattr(self, 'f%d' % i)
            # 2D
            elif len(shape) == 2:
                for i in range(shape[0]):
                    for j in range(shape[1]):
                        value[i, j] = getattr(self, 'f%d_%d' % (i, j))

            self.editor.update_array(value)


#-------------------------------------------------------------------------
#  Toolkit-independent 'SimpleEditor' class:
#-------------------------------------------------------------------------

class SimpleEditor(Editor):
    """ Simple style of editor for arrays.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Is the editor read-only?
    readonly = false

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self._as = _as = ArrayStructure(self)
        ui = _as.view.ui(_as, parent, kind='subpanel')
        ui.parent = self.ui
        self.control = ui.control

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """

        if not self._busy:
            self._busy = True
            object = self.value
            shape = object.shape
            _as = self._as

            # 1D
            if len(shape) == 1:
                for i in range(shape[0]):
                    setattr(_as, 'f%d' % i, object[i])
            # 2D
            elif len(shape) == 2:
                for i in range(shape[0]):
                    for j in range(shape[1]):
                        setattr(_as, 'f%d_%d' % (i, j), object[i, j])

            self._busy = False

    #-------------------------------------------------------------------------
    #  Updates the array value associated with the editor:
    #-------------------------------------------------------------------------

    def update_array(self, value):
        """ Updates the array value associated with the editor.
        """
        self._busy = True
        self.value = value
        self._busy = False


# Define the ArrayEditor class
ArrayEditor = ToolkitEditorFactory

### EOF ---------------------------------------------------------------------



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The interface for a dialog that allows the user to browse for a directory.
"""


# Enthought library imports.
from traits.api import Bool, Unicode

# Local imports.
from pyface.i_dialog import IDialog


class IDirectoryDialog(IDialog):
    """ The interface for a dialog that allows the user to browse for a
    directory.
    """

    #### 'IDirectoryDialog' interface #########################################

    #: The default path.  The default (ie. the default default path) is toolkit
    #: specific.
    # FIXME v3: The default should be the current directory.  (It seems wx is
    # the problem, although the file dialog does the right thing.)
    default_path = Unicode

    #: The message to display in the dialog.  The default is toolkit specific.
    message = Unicode

    #: True iff the dialog should include a button that allows the user to
    #: create a new directory.
    new_directory = Bool(True)

    #: The path of the chosen directory.
    path = Unicode


class MDirectoryDialog(object):
    """ The mixin class that contains common code for toolkit specific
    implementations of the IDirectoryDialog interface.
    """



""" Html Viewer

"""
# Enthought library imports.
from pyface.api import GUI, ImageResource, ConfirmationDialog, FileDialog, \
    ImageResource, YES, OK, CANCEL
from pyface.action.api import Action, ActionItem, Separator, Group
from pyface.tasks.api import Task, TaskWindow, TaskLayout, PaneItem, IEditor, \
    IEditorAreaPane, EditorAreaPane, Editor, DockPane, HSplitter, VSplitter
from pyface.tasks.action.api import DockPaneToggleGroup, SMenuBar, \
    SMenu, SToolBar, TaskAction, TaskToggleGroup, EditorAction, SchemaAddition
from traits.api import on_trait_change, Property, Instance, Any, Event, Int

from omnivore.framework.task import FrameworkTask
from omnivore.framework.actions import *
from html_viewer import HtmlViewer
from preferences import HtmlViewPreferences
import pane_layout


class HtmlViewTask(FrameworkTask):
    """ Static HTML viewer
    """

    new_file_text = ""

    editor_id = "omnivore.htmlview"

    pane_layout_version = pane_layout.pane_layout_version

    doc_hint = "skip"

    #### Task interface #######################################################

    id = editor_id
    name = 'HTML Viewer'

    preferences_helper = HtmlViewPreferences

    ###########################################################################
    # 'Task' interface.
    ###########################################################################

    def _default_layout_default(self):
        return pane_layout.pane_layout()

    def create_dock_panes(self):
        return pane_layout.pane_create()

    ###########################################################################
    # 'FrameworkTask' interface.
    ###########################################################################

    def get_editor(self, guess=None):
        """ Opens a new empty window
        """
        editor = HtmlViewer()
        return editor

    ###
    @classmethod
    def can_edit(cls, document):
        return document.metadata.mime == "text/html"

    @classmethod
    def get_match_score(cls, guess):
        """Return a number based on how good of a match this task is to the
        incoming FileGuess.
        
        0 = generic match
        ...
        10 = absolute match
        """
        if guess.metadata.mime == "text/html":
            return 10
        return 1



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#  fixme: Add undo/redo support
#  fixme: Allow factory to handle a TraitListObject for the 'values' trait.
#
#------------------------------------------------------------------------------

""" Defines the set editors for the wxPython user interface toolkit.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import Property

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.set_editor file.
from traitsui.editors.set_editor \
    import ToolkitEditorFactory

from editor \
    import Editor

from helper \
    import enum_values_changed, TraitsUIPanel

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class SimpleEditor(Editor):
    """ Simple style of editor for sets.

        The editor displays two list boxes, with buttons for moving the selected
        items from left to right, or vice versa. If **can_move_all** on the
        factory is True, then buttons are displayed for moving all the items to
        one box or the other. If the set is ordered, buttons are displayed for
        moving the selected item up or down in right-side list box.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Current set of enumeration names:
    names = Property

    # Current mapping from names to values:
    mapping = Property

    # Current inverse mapping from values to names:
    inverse_mapping = Property

    # Is set editor scrollable? This value overrides the default.
    scrollable = True

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory
        if factory.name != '':
            self._object, self._name, self._value = \
                self.parse_extended_name(factory.name)
            self.values_changed()
            self._object.on_trait_change(self._values_changed,
                                         self._name, dispatch='ui')
        else:
            factory.on_trait_change(self.update_editor, 'values_modified',
                                    dispatch='ui')

        self.control = panel = TraitsUIPanel(parent, -1)
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        vsizer = wx.BoxSizer(wx.VERTICAL)

        self._unused = self._create_listbox(
            panel,
            hsizer,
            self._on_unused,
            self._on_use,
            factory.left_column_title)

        self._use_all = self._unuse_all = self._up = self._down = None
        if factory.can_move_all:
            self._use_all = self._create_button('>>', panel, vsizer, 15,
                                                self._on_use_all)

        self._use = self._create_button('>', panel, vsizer, 15,
                                        self._on_use)
        self._unuse = self._create_button('<', panel, vsizer, 0,
                                          self._on_unuse)
        if factory.can_move_all:
            self._unuse_all = self._create_button('<<', panel, vsizer, 15,
                                                  self._on_unuse_all)

        if factory.ordered:
            self._up = self._create_button('Move Up', panel, vsizer, 30,
                                           self._on_up)
            self._down = self._create_button('Move Down', panel, vsizer, 0,
                                             self._on_down)

        hsizer.Add(vsizer, 0, wx.LEFT | wx.RIGHT, 8)
        self._used = self._create_listbox(
            panel,
            hsizer,
            self._on_value,
            self._on_unuse,
            factory.right_column_title)

        panel.SetSizer(hsizer)

        self.context_object.on_trait_change(
            self.update_editor,
            self.extended_name + '_items?',
            dispatch='ui')
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Gets the current set of enumeration names:
    #-------------------------------------------------------------------------

    def _get_names(self):
        """ Gets the current set of enumeration names.
        """
        if self._object is None:
            return self.factory._names

        return self._names

    #-------------------------------------------------------------------------
    #  Gets the current mapping:
    #-------------------------------------------------------------------------

    def _get_mapping(self):
        """ Gets the current mapping.
        """
        if self._object is None:
            return self.factory._mapping

        return self._mapping

    #-------------------------------------------------------------------------
    #  Gets the current inverse mapping:
    #-------------------------------------------------------------------------

    def _get_inverse_mapping(self):
        """ Gets the current inverse mapping.
        """
        if self._object is None:
            return self.factory._inverse_mapping

        return self._inverse_mapping

    #-------------------------------------------------------------------------
    #  Creates a list box:
    #-------------------------------------------------------------------------

    def _create_listbox(self, parent, sizer, handler1, handler2, title):
        """ Creates a list box.
        """
        column_sizer = wx.BoxSizer(wx.VERTICAL)

        # Add the column title in emphasized text:
        title_widget = wx.StaticText(parent, -1, title)
        font = title_widget.GetFont()
        emphasis_font = wx.Font(font.GetPointSize() + 1,
                                font.GetFamily(),
                                font.GetStyle(),
                                wx.BOLD)
        title_widget.SetFont(emphasis_font)
        column_sizer.Add(title_widget, 0, 0)

        # Create the list box and add it to the column:
        list = wx.ListBox(parent, -1, style=wx.LB_EXTENDED | wx.LB_NEEDED_SB)
        column_sizer.Add(list, 1, wx.EXPAND)

        # Add the column to the SetEditor widget:
        sizer.Add(column_sizer, 1, wx.EXPAND)

        # Hook up the event handlers:
        wx.EVT_LISTBOX(parent, list.GetId(), handler1)
        wx.EVT_LISTBOX_DCLICK(parent, list.GetId(), handler2)

        return list

    #-------------------------------------------------------------------------
    #  Creates a button:
    #-------------------------------------------------------------------------

    def _create_button(self, label, parent, sizer, space_before, handler):
        """ Creates a button.
        """
        button = wx.Button(parent, -1, label, style=wx.BU_EXACTFIT)
        sizer.AddSpacer((space_before, space_before))
        sizer.Add(button, 0, wx.EXPAND | wx.BOTTOM, 8)
        wx.EVT_BUTTON(parent, button.GetId(), handler)
        return button

    #-------------------------------------------------------------------------
    #  Recomputes the cached data based on the underlying enumeration model:
    #-------------------------------------------------------------------------

    def values_changed(self):
        """ Recomputes the cached data based on the underlying enumeration model.
        """
        self._names, self._mapping, self._inverse_mapping = \
            enum_values_changed(self._value())

    #-------------------------------------------------------------------------
    #  Handles the underlying object model's enumeration set being changed:
    #-------------------------------------------------------------------------

    def _values_changed(self):
        """ Handles the underlying object model's enumeration set being changed.
        """
        self.values_changed()
        self.update_editor()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        # Check for any items having been deleted from the enumeration that are
        # still present in the object value:
        mapping = self.inverse_mapping.copy()
        values = [v for v in self.value if v in mapping]
        if len(values) < len(self.value):
            self.value = values
            return

        # Get a list of the selected items in the right box:
        used = self._used
        used_labels = self._get_selected_strings(used)

        # Get a list of the selected items in the left box:
        unused = self._unused
        unused_labels = self._get_selected_strings(unused)

        # Empty list boxes in preparation for rebuilding from current values:
        used.Clear()
        unused.Clear()

        # Ensure right list box is kept alphabetized unless insertion
        # order is relevant:
        if not self.factory.ordered:
            values = sorted(values[:])

        # Rebuild the right listbox:
        used_selections = []
        for i, value in enumerate(values):
            label = mapping[value]
            used.Append(label)
            del mapping[value]
            if label in used_labels:
                used_selections.append(i)

        # Rebuild the left listbox:
        unused_selections = []
        unused_items = sorted(mapping.values())
        mapping = self.mapping
        self._unused_items = [mapping[ui] for ui in unused_items]
        for i, unused_item in enumerate(unused_items):
            unused.Append(unused_item)
            if unused_item in unused_labels:
                unused_selections.append(i)

        # If nothing is selected, default selection should be top of left box,
        # or of right box if left box is empty:
        if (len(used_selections) == 0) and (len(unused_selections) == 0):
            if unused.GetCount() == 0:
                used_selections.append(0)
            else:
                unused_selections.append(0)

        used_count = used.GetCount()
        for i in used_selections:
            if i < used_count:
                used.SetSelection(i)

        unused_count = unused.GetCount()
        for i in unused_selections:
            if i < unused_count:
                unused.SetSelection(i)

        self._check_up_down()
        self._check_left_right()

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        if self._object is not None:
            self._object.on_trait_change(self._values_changed,
                                         self._name, remove=True)
        else:
            self.factory.on_trait_change(self.update_editor,
                                         'values_modified', remove=True)

        self.context_object.on_trait_change(
            self.update_editor,
            self.extended_name + '_items?',
            remove=True)

        super(SimpleEditor, self).dispose()

    #-------------------------------------------------------------------------
    #  Returns the editor's control for indicating error status:
    #-------------------------------------------------------------------------

    def get_error_control(self):
        """ Returns the editor's control for indicating error status.
        """
        return [self._unused, self._used]

    #-------------------------------------------------------------------------
    #  Event handlers:
    #-------------------------------------------------------------------------

    def _on_value(self, event):
        if not self.factory.ordered:
            self._clear_selection(self._unused)
        self._check_left_right()
        self._check_up_down()

    def _on_unused(self, event):
        if not self.factory.ordered:
            self._clear_selection(self._used)
        self._check_left_right()
        self._check_up_down()

    def _on_use(self, event):
        self._unused_items, self.value = self._transfer_items(
            self._unused, self._used, self._unused_items, self.value)

    def _on_unuse(self, event):
        self.value, self._unused_items = self._transfer_items(
            self._used, self._unused, self.value, self._unused_items)

    def _on_use_all(self, event):
        self._unused_items, self.value = self._transfer_all(
            self._unused, self._used, self._unused_items, self.value)

    def _on_unuse_all(self, event):
        self.value, self._unused_items = self._transfer_all(
            self._used, self._unused, self.value, self._unused_items)

    def _on_up(self, event):
        self._move_item(-1)

    def _on_down(self, event):
        self._move_item(1)

    #-------------------------------------------------------------------------
    #  Private methods:
    #-------------------------------------------------------------------------

    #-------------------------------------------------------------------------
    # Unselects all items in the given ListBox
    #-------------------------------------------------------------------------

    def _clear_selection(self, box):
        """ Unselects all items in the given ListBox
        """
        for i in box.GetSelections():
            box.SetSelection(i, False)

    #-------------------------------------------------------------------------
    #  Transfers all items from one list to another:
    #-------------------------------------------------------------------------

    def _transfer_all(self, list_from, list_to, values_from, values_to):
        """ Transfers all items from one list to another.
        """
        values_from = values_from[:]
        values_to = values_to[:]

        self._clear_selection(list_from)
        while list_from.GetCount() > 0:
            index_to = list_to.GetCount()
            list_from.SetSelection(0)
            list_to.InsertItems(self._get_selected_strings(list_from),
                                index_to)
            list_from.Delete(0)
            values_to.append(values_from[0])
            del values_from[0]

        list_to.SetSelection(0)
        self._check_left_right()
        self._check_up_down()

        return (values_from, values_to)

    #-------------------------------------------------------------------------
    #  Transfers the selected item from one list to another:
    #-------------------------------------------------------------------------

    def _transfer_items(self, list_from, list_to, values_from, values_to):
        """ Transfers the selected item from one list to another.
        """
        values_from = values_from[:]
        values_to = values_to[:]
        indices_from = list_from.GetSelections()
        index_from = max(self._get_first_selection(list_from), 0)
        index_to = max(self._get_first_selection(list_to), 0)

        self._clear_selection(list_to)

        # Get the list of strings in the "from" box to be moved:
        selected_list = self._get_selected_strings(list_from)

        # fixme: I don't know why I have to reverse the list to get
        # correct behavior from the ordered list box.  Investigate -- LP
        selected_list.reverse()
        list_to.InsertItems(selected_list, index_to)

        # Delete the transferred items from the left box:
        for i in range(len(indices_from) - 1, -1, -1):
            list_from.Delete(indices_from[i])

        # Delete the transferred items from the "unused" value list:
        for item_label in selected_list:
            val_index_from = values_from.index(self.mapping[item_label])
            values_to.insert(index_to, values_from[val_index_from])
            del values_from[val_index_from]

            # If right list is ordered, keep moved items selected:
            if self.factory.ordered:
                list_to.SetSelection(list_to.FindString(item_label))

        # Reset the selection in the left box:
        count = list_from.GetCount()
        if count > 0:
            if index_from >= count:
                index_from -= 1
            list_from.SetSelection(index_from)

        self._check_left_right()
        self._check_up_down()

        return (values_from, values_to)

    #-------------------------------------------------------------------------
    #  Moves an item up or down with the 'used' list:
    #-------------------------------------------------------------------------

    def _move_item(self, direction):
        """ Moves an item up or down within the "used" list.
        """
        # Move the item up/down within the list:
        listbox = self._used
        index_from = self._get_first_selection(listbox)
        index_to = index_from + direction
        label = listbox.GetString(index_from)
        listbox.Delete(index_from)
        listbox.Insert(label, index_to)
        listbox.SetSelection(index_to)

        # Enable the up/down buttons appropriately:
        self._check_up_down()

        # Move the item up/down within the editor's trait value:
        value = self.value
        if direction < 0:
            index = index_to
            values = [value[index_from], value[index_to]]
        else:
            index = index_from
            values = [value[index_to], value[index_from]]
        self.value = value[: index] + values + value[index + 2:]

    #-------------------------------------------------------------------------
    #  Sets the proper enable state for the up and down buttons:
    #-------------------------------------------------------------------------

    def _check_up_down(self):
        """ Sets the proper enabled state for the up and down buttons.
        """
        if self.factory.ordered:
            index_selected = self._used.GetSelections()
            self._up.Enable((len(index_selected) == 1) and
                            (index_selected[0] > 0))
            self._down.Enable(
                (len(index_selected) == 1) and (
                    index_selected[0] < (
                        self._used.GetCount() - 1)))

    #-------------------------------------------------------------------------
    #  Sets the proper enable state for the left and right buttons:
    #-------------------------------------------------------------------------

    def _check_left_right(self):
        """ Sets the proper enabled state for the left and right buttons.
        """
        self._use.Enable(self._unused.GetCount() > 0 and
                         self._get_first_selection(self._unused) >= 0)
        self._unuse.Enable(self._used.GetCount() > 0 and
                           self._get_first_selection(self._used) >= 0)

        if self.factory.can_move_all:
            self._use_all.Enable(
                (self._unused.GetCount() > 0) and (
                    self._get_first_selection(
                        self._unused) >= 0))
            self._unuse_all.Enable(
                (self._used.GetCount() > 0) and (
                    self._get_first_selection(
                        self._used) >= 0))

    #-------------------------------------------------------------------------
    # Returns a list of the selected strings in the listbox
    #-------------------------------------------------------------------------

    def _get_selected_strings(self, listbox):
        """ Returns a list of the selected strings in the given *listbox*.
        """
        stringlist = []
        for label_index in listbox.GetSelections():
            stringlist.append(listbox.GetString(label_index))

        return stringlist

    #-------------------------------------------------------------------------
    # Returns the index of the first (or only) selected item.
    #-------------------------------------------------------------------------

    def _get_first_selection(self, listbox):
        """ Returns the index of the first (or only) selected item.
        """
        select_list = listbox.GetSelections()
        if len(select_list) == 0:
            return -1

        return select_list[0]

### EOF #######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought IO package component>
#------------------------------------------------------------------------------
""" A representation of files and folders in a file system. """


# Standard/built-in imports.
import mimetypes, os, shutil, stat

# Enthought library imports.
from traits.api import Bool, HasPrivateTraits, Instance, List, Property
from traits.api import Str


class File(HasPrivateTraits):
    """ A representation of files and folders in a file system. """

    #### 'File' interface #####################################################

    # The absolute path name of this file/folder.
    absolute_path = Property(Str)

    # The folder's children (for files this is always None).
    children = Property(List('File'))

    # The file extension (for folders this is always the empty string).
    #
    # fixme: Currently the extension includes the '.' (ie. we have '.py' and
    # not 'py'). This is because things like 'os.path.splitext' leave the '.'
    # on, but I'm not sure that this is a good idea!
    ext = Property(Str)

    # Does the file/folder exist?
    exists = Property(Bool)

    # Is this an existing file?
    is_file = Property(Bool)

    # Is this an existing folder?
    is_folder = Property(Bool)

    # Is this a Python package (ie. a folder contaning an '__init__.py' file.
    is_package = Property(Bool)

    # Is the file/folder readonly?
    is_readonly = Property(Bool)

    # The MIME type of the file (for a folder this will always be
    # 'context/unknown' (is that what it should be?)).
    mime_type = Property(Str)

    # The last component of the path without the extension.
    name = Property(Str)

    # The parent of this file/folder (None if it has no parent).
    parent = Property(Instance('File'))

    # The path name of this file/folder.
    path = Str

    # A URL reference to the file.
    url = Property(Str)

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, path, **traits):
        """ Creates a new representation of the specified path. """

        super(File, self).__init__(path=path, **traits)

        return

    def __cmp__(self, other):
        """ Comparison operators. """
        if isinstance(other, File):
            return cmp(self.path, other.path)

        return 1

    def __str__(self):
        """ Returns an 'informal' string representation of the object. """

        return 'File(%s)' % self.path

    ###########################################################################
    # 'File' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get_absolute_path(self):
        """ Returns the absolute path of this file/folder. """

        return os.path.abspath(self.path)

    def _get_children(self):
        """ Returns the folder's children.

        Returns None if the path does not exist or is not a folder.

        """

        if self.is_folder:
            children = []
            for name in os.listdir(self.path):
                children.append(File(os.path.join(self.path, name)))

        else:
            children = None

        return children

    def _get_exists(self):
        """ Returns True if the file exists, otherwise False. """

        return os.path.exists(self.path)

    def _get_ext(self):
        """ Returns the file extension. """

        name, ext = os.path.splitext(self.path)

        return ext

    def _get_is_file(self):
        """ Returns True if the path exists and is a file. """

        return self.exists and os.path.isfile(self.path)

    def _get_is_folder(self):
        """ Returns True if the path exists and is a folder. """

        return self.exists and os.path.isdir(self.path)

    def _get_is_package(self):
        """ Returns True if the path exists and is a Python package. """

        return self.is_folder and '__init__.py' in os.listdir(self.path)

    def _get_is_readonly(self):
        """ Returns True if the file/folder is readonly, otherwise False. """

        # If the File object is a folder, os.access cannot be used because it
        # returns True for both read-only and writable folders on Windows
        # systems.
        if self.is_folder:

            # Mask for the write-permission bits on the folder. If these bits
            # are set to zero, the folder is read-only.
            WRITE_MASK = 0x92
            permissions = os.stat(self.path)[0]

            if permissions & WRITE_MASK == 0:
                readonly = True
            else:
                readonly = False

        elif self.is_file:
            readonly = not os.access(self.path, os.W_OK)

        else:
            readonly = False

        return readonly

    def _get_mime_type(self):
        """ Returns the mime-type of this file/folder. """

        mime_type, encoding = mimetypes.guess_type(self.path)
        if mime_type is None:
            mime_type = "content/unknown"

        return mime_type

    def _get_name(self):
        """ Returns the last component of the path without the extension. """

        basename = os.path.basename(self.path)

        name, ext = os.path.splitext(basename)

        return name

    def _get_parent(self):
        """ Returns the parent of this file/folder. """

        return File(os.path.dirname(self.path))

    def _get_url(self):
        """ Returns the path as a URL. """

        # Strip out the leading slash on POSIX systems.
        return 'file:///%s' % self.absolute_path.lstrip('/')

    #### Methods ##############################################################

    def copy(self, destination):
        """ Copies this file/folder. """

        # Allow the destination to be a string.
        if not isinstance(destination, File):
            destination = File(destination)

        if self.is_folder:
            shutil.copytree(self.path, destination.path)

        elif self.is_file:
            shutil.copyfile(self.path, destination.path)

        return

    def create_file(self, contents=''):
        """ Creates a file at this path. """

        if self.exists:
            raise ValueError("file %s already exists" % self.path)

        f = open(self.path, 'w')
        f.write(contents)
        f.close()

        return

    def create_folder(self):
        """ Creates a folder at this path.

        All intermediate folders MUST already exist.

        """

        if self.exists:
            raise ValueError("folder %s already exists" % self.path)

        os.mkdir(self.path)

        return

    def create_folders(self):
        """ Creates a folder at this path.

        This will attempt to create any missing intermediate folders.

        """

        if self.exists:
            raise ValueError("folder %s already exists" % self.path)

        os.makedirs(self.path)

        return

    def create_package(self):
        """ Creates a package at this path.

        All intermediate folders/packages MUST already exist.

        """

        if self.exists:
            raise ValueError("package %s already exists" % self.path)

        os.mkdir(self.path)

        # Create the '__init__.py' file that actually turns the folder into a
        # package!
        init = File(os.path.join(self.path, '__init__.py'))
        init.create_file()

        return

    def delete(self):
        """ Deletes this file/folder.

        Does nothing if the file/folder does not exist.

        """

        if self.is_folder:
            # Try to make sure that everything in the folder is writeable.
            self.make_writeable()

            # Delete it!
            shutil.rmtree(self.path)

        elif self.is_file:
            # Try to make sure that the file is writeable.
            self.make_writeable()

            # Delete it!
            os.remove(self.path)

        return

    def make_writeable(self):
        """ Attempt to make the file/folder writeable. """

        if self.is_folder:
            # Try to make sure that everything in the folder is writeable
            # (i.e., can be deleted!).  This comes in especially handy when
            # deleting '.svn' directories.
            for path, dirnames, filenames in os.walk(self.path):
                for name in dirnames + filenames:
                    filename = os.path.join(path, name)
                    if not os.access(filename, os.W_OK):
                        os.chmod(filename, stat.S_IWUSR)

        elif self.is_file:
            # Try to make sure that the file is writeable (i.e., can be
            # deleted!).
            if not os.access(self.path, os.W_OK):
                os.chmod(self.path, stat.S_IWUSR)

        return

    def move(self, destination):
        """ Moves this file/folder. """

        # Allow the destination to be a string.
        if not isinstance(destination, File):
            destination = File(destination)

        # Try to make sure that everything in the directory is writeable.
        self.make_writeable()

        # Move it!
        shutil.move(self.path, destination.path)

        return

#### EOF ######################################################################



"""
fs.remote
=========

Utilities for interfacing with remote filesystems


This module provides reusable utility functions that can be used to construct
FS subclasses interfacing with a remote filesystem.  These include:

  * RemoteFileBuffer:  a file-like object that locally buffers the contents of
                       a remote file, writing them back on flush() or close().

  * ConnectionManagerFS:  a WrapFS subclass that tracks the connection state
                          of a remote FS, and allows client code to wait for
                          a connection to be re-established.

  * CacheFS:  a WrapFS subclass that caches file and directory meta-data in
              memory, to speed access to a remote FS.

"""

from __future__ import with_statement

import time
import stat as statinfo
from errno import EINVAL

import fs.utils
from fs.base import threading, FS
from fs.wrapfs import WrapFS, wrap_fs_methods
from fs.wrapfs.lazyfs import LazyFS
from fs.path import *
from fs.errors import *
from fs.local_functools import wraps
from fs.filelike import StringIO, SpooledTemporaryFile, FileWrapper
from fs import SEEK_SET, SEEK_CUR, SEEK_END


_SENTINAL = object()

from six import PY3, b


class RemoteFileBuffer(FileWrapper):
    """File-like object providing buffer for local file operations.

    Instances of this class manage a local tempfile buffer corresponding
    to the contents of a remote file.  All reads and writes happen locally,
    with the content being copied to the remote file only on flush() or
    close().  Writes to the remote file are performed using the setcontents()
    method on the owning FS object.

    The intended use-case is for a remote filesystem (e.g. S3FS) to return
    instances of this class from its open() method, and to provide the
    file-uploading logic in its setcontents() method, as in the following
    pseudo-code::

        def open(self,path,mode="r"):
            rf = self._get_remote_file(path)
            return RemoteFileBuffer(self,path,mode,rf)

        def setcontents(self,path,file):
            self._put_remote_file(path,file)

    The contents of the remote file are read into the buffer on-demand.
    """

    max_size_in_memory = 1024 * 8

    def __init__(self, fs, path, mode, rfile=None, write_on_flush=True):
        """RemoteFileBuffer constructor.

        The owning filesystem, path and mode must be provided.  If the
        optional argument 'rfile' is provided, it must be a read()-able
        object or a string containing the initial file contents.
        """
        wrapped_file = SpooledTemporaryFile(max_size=self.max_size_in_memory)
        self.fs = fs
        self.path = path
        self.write_on_flush = write_on_flush
        self._changed = False
        self._readlen = 0  # How many bytes already loaded from rfile
        self._rfile = None  # Reference to remote file object
        self._eof = False  # Reached end of rfile?
        if getattr(fs, "_lock", None) is not None:
            self._lock = fs._lock.__class__()
        else:
            self._lock = threading.RLock()

        if "r" in mode or "+" in mode or "a" in mode:
            if rfile is None:
                # File was just created, force to write anything
                self._changed = True
                self._eof = True

            if not hasattr(rfile, "read"):
                #rfile = StringIO(unicode(rfile))
                rfile = StringIO(rfile)

            self._rfile = rfile
        else:
            # Do not use remote file object
            self._eof = True
            self._rfile = None
            self._changed = True
            if rfile is not None and hasattr(rfile,"close"):
                rfile.close()
        super(RemoteFileBuffer,self).__init__(wrapped_file,mode)
        # FIXME: What if mode with position on eof?
        if "a" in mode:
            # Not good enough...
            self.seek(0, SEEK_END)

    def __del__(self):
        #  Don't try to close a partially-constructed file
        if "_lock" in self.__dict__:
            if not self.closed:
                try:
                    self.close()
                except FSError:
                    pass

    def _write(self,data,flushing=False):
        with self._lock:
            #  Do we need to discard info from the buffer?
            toread = len(data) - (self._readlen - self.wrapped_file.tell())
            if toread > 0:
                if not self._eof:
                    self._fillbuffer(toread)
                else:
                    self._readlen += toread
            self._changed = True
            self.wrapped_file.write(data)

    def _read_remote(self, length=None):
        """Read data from the remote file into the local buffer."""
        chunklen = 1024 * 256
        bytes_read = 0
        while True:
            toread = chunklen
            if length is not None and length - bytes_read < chunklen:
                toread = length - bytes_read
            if not toread:
                break

            data = self._rfile.read(toread)
            datalen = len(data)
            if not datalen:
                self._eof = True
                break

            bytes_read += datalen
            self.wrapped_file.write(data)

            if datalen < toread:
                # We reached EOF,
                # no more reads needed
                self._eof = True
                break

        if self._eof and self._rfile is not None:
            self._rfile.close()
        self._readlen += bytes_read

    def _fillbuffer(self, length=None):
        """Fill the local buffer, leaving file position unchanged.

        This method is used for on-demand loading of data from the remote file
        into the buffer.  It reads 'length' bytes from rfile and writes them
        into the buffer, seeking back to the original file position.
        """
        curpos = self.wrapped_file.tell()
        if length == None:
            if not self._eof:
                # Read all data and we didn't reached EOF
                # Merge endpos - tell + bytes from rfile
                self.wrapped_file.seek(0, SEEK_END)
                self._read_remote()
                self._eof = True
                self.wrapped_file.seek(curpos)

        elif not self._eof:
            if curpos + length > self._readlen:
                # Read all data and we didn't reached EOF
                # Load endpos - tell() + len bytes from rfile
                toload = length - (self._readlen - curpos)
                self.wrapped_file.seek(0, SEEK_END)
                self._read_remote(toload)
                self.wrapped_file.seek(curpos)

    def _read(self, length=None):
        if length is not None and length < 0:
            length = None
        with self._lock:
            self._fillbuffer(length)
            data = self.wrapped_file.read(length if length != None else -1)
            if not data:
                data = None
            return data

    def _seek(self,offset,whence=SEEK_SET):
        with self._lock:
            if not self._eof:
                # Count absolute position of seeking
                if whence == SEEK_SET:
                    abspos = offset
                elif whence == SEEK_CUR:
                    abspos =  offset + self.wrapped_file.tell()
                elif whence == SEEK_END:
                    abspos = None
                else:
                    raise IOError(EINVAL, 'Invalid whence')

                if abspos != None:
                    toread = abspos - self._readlen
                    if toread > 0:
                        self.wrapped_file.seek(self._readlen)
                        self._fillbuffer(toread)
                else:
                    self.wrapped_file.seek(self._readlen)
                    self._fillbuffer()

            self.wrapped_file.seek(offset, whence)

    def _truncate(self,size):
        with self._lock:
            if not self._eof and self._readlen < size:
                # Read the rest of file
                self._fillbuffer(size - self._readlen)
                # Lock rfile
                self._eof = True
            elif self._readlen >= size:
                # Crop rfile metadata
                self._readlen = size if size != None else 0
                # Lock rfile
                self._eof = True

            self.wrapped_file.truncate(size)
            self._changed = True

            self.flush()
            if self._rfile is not None:
                self._rfile.close()

    def flush(self):
        with self._lock:
            self.wrapped_file.flush()
            if self.write_on_flush:
                self._setcontents()

    def _setcontents(self):
        if not self._changed:
            # Nothing changed, no need to write data back
            return

        # If not all data loaded, load until eof
        if not self._eof:
            self._fillbuffer()

        if "w" in self.mode or "a" in self.mode or "+" in self.mode:
            pos = self.wrapped_file.tell()
            self.wrapped_file.seek(0)
            self.fs.setcontents(self.path, self.wrapped_file)
            self.wrapped_file.seek(pos)

    def close(self):
        with self._lock:
            if not self.closed:
                self._setcontents()
                if self._rfile is not None:
                    self._rfile.close()
                super(RemoteFileBuffer,self).close()


class ConnectionManagerFS(LazyFS):
    """FS wrapper providing simple connection management of a remote FS.

    The ConnectionManagerFS class is designed to wrap a remote FS object
    and provide some convenience methods for dealing with its remote
    connection state.

    The boolean attribute 'connected' indicates whether the remote filesystem
    has an active connection, and is initially True.  If any of the remote
    filesystem methods raises a RemoteConnectionError, 'connected' will
    switch to False and remain so until a successful remote method call.

    Application code can use the method 'wait_for_connection' to block
    until the connection is re-established.  Currently this reconnection
    is checked by a simple polling loop; eventually more sophisticated
    operating-system integration may be added.

    Since some remote FS classes can raise RemoteConnectionError during
    initialization, this class makes use of lazy initialization. The
    remote FS can be specified as an FS instance, an FS subclass, or a
    (class,args) or (class,args,kwds) tuple. For example::

        >>> fs = ConnectionManagerFS(MyRemoteFS("http://www.example.com/"))
        Traceback (most recent call last):
            ...
        RemoteConnectionError: couldn't connect to "http://www.example.com/"
        >>> fs = ConnectionManagerFS((MyRemoteFS,["http://www.example.com/"]))
        >>> fs.connected
        False
        >>>

    """

    poll_interval = 1

    def __init__(self,wrapped_fs,poll_interval=None,connected=True):
        super(ConnectionManagerFS,self).__init__(wrapped_fs)
        if poll_interval is not None:
            self.poll_interval = poll_interval
        self._connection_cond = threading.Condition()
        self._poll_thread = None
        self._poll_sleeper = threading.Event()
        self.connected = connected

    def setcontents(self, path, data=b'', encoding=None, errors=None, chunk_size=64*1024):
        return self.wrapped_fs.setcontents(path, data, encoding=encoding, errors=errors, chunk_size=chunk_size)

    def __getstate__(self):
        state = super(ConnectionManagerFS,self).__getstate__()
        del state["_connection_cond"]
        del state["_poll_sleeper"]
        state["_poll_thread"] = None
        return state

    def __setstate__(self,state):
        super(ConnectionManagerFS,self).__setstate__(state)
        self._connection_cond = threading.Condition()
        self._poll_sleeper = threading.Event()

    def wait_for_connection(self,timeout=None,force_wait=False):
        self._connection_cond.acquire()
        try:
            if force_wait:
                self.connected = False
            if not self.connected:
                if not self._poll_thread:
                    target = self._poll_connection
                    self._poll_thread = threading.Thread(target=target)
                    self._poll_thread.daemon = True
                    self._poll_thread.start()
                self._connection_cond.wait(timeout)
        finally:
            self._connection_cond.release()

    def _poll_connection(self):
        while not self.connected and not self.closed:
            try:
                self.wrapped_fs.getinfo("/")
            except RemoteConnectionError:
                self._poll_sleeper.wait(self.poll_interval)
                self._poll_sleeper.clear()
            except FSError:
                break
            else:
                break
        self._connection_cond.acquire()
        try:
            if not self.closed:
                self.connected = True
            self._poll_thread = None
            self._connection_cond.notifyAll()
        finally:
            self._connection_cond.release()

    def close(self):
        if not self.closed:
            try:
                super(ConnectionManagerFS,self).close()
            except (RemoteConnectionError,):
                pass
            if self._poll_thread:
                self.connected = True
                self._poll_sleeper.set()
                self._poll_thread.join()
                self._poll_thread = None

def _ConnectionManagerFS_method_wrapper(func):
    """Method wrapper for ConnectionManagerFS.

    This method wrapper keeps an eye out for RemoteConnectionErrors and
    adjusts self.connected accordingly.
    """
    @wraps(func)
    def wrapper(self,*args,**kwds):
        try:
            result = func(self,*args,**kwds)
        except RemoteConnectionError:
            self.connected = False
            raise
        except FSError:
            self.connected = True
            raise
        else:
            self.connected = True
            return result
    return wrapper

wrap_fs_methods(_ConnectionManagerFS_method_wrapper,ConnectionManagerFS)


class CachedInfo(object):
    """Info objects stored in cache for CacheFS."""
    __slots__ = ("timestamp","info","has_full_info","has_full_children")
    def __init__(self,info={},has_full_info=True,has_full_children=False):
        self.timestamp = time.time()
        self.info = info
        self.has_full_info = has_full_info
        self.has_full_children = has_full_children
    def clone(self):
        new_ci = self.__class__()
        new_ci.update_from(self)
        return new_ci
    def update_from(self,other):
        self.timestamp = other.timestamp
        self.info = other.info
        self.has_full_info = other.has_full_info
        self.has_full_children = other.has_full_children
    @classmethod
    def new_file_stub(cls):
        info = {"info" : 0700 | statinfo.S_IFREG}
        return cls(info,has_full_info=False)
    @classmethod
    def new_dir_stub(cls):
        info = {"info" : 0700 | statinfo.S_IFDIR}
        return cls(info,has_full_info=False)


class CacheFSMixin(FS):
    """Simple FS mixin to cache meta-data of a remote filesystems.

    This FS mixin implements a simplistic cache that can help speed up
    access to a remote filesystem.  File and directory meta-data is cached
    but the actual file contents are not.

    If you want to add caching to an existing FS object, use the CacheFS
    class instead; it's an easy-to-use wrapper rather than a mixin.
    This mixin class is provided for FS implementors who want to use
    caching internally in their own classes.

    FYI, the implementation of CacheFS is this:

        class CacheFS(CacheFSMixin,WrapFS):
            pass

    """

    def __init__(self,*args,**kwds):
        """CacheFSMixin constructor.

        The optional keyword argument 'cache_timeout' specifies the cache
        timeout in seconds.  The default timeout is 1 second.  To prevent
        cache entries from ever timing out, set it to None.

        The optional keyword argument 'max_cache_size' specifies the maximum
        number of entries to keep in the cache.  To allow the cache to grow
        without bound, set it to None.  The default is 1000.
        """
        self.cache_timeout = kwds.pop("cache_timeout",1)
        self.max_cache_size = kwds.pop("max_cache_size",1000)
        self.__cache = PathMap()
        self.__cache_size = 0
        self.__cache_lock = threading.RLock()
        super(CacheFSMixin,self).__init__(*args,**kwds)

    def clear_cache(self,path=""):
        with self.__cache_lock:
            self.__cache.clear(path)
        try:
            scc = super(CacheFSMixin,self).clear_cache
        except AttributeError:
            pass
        else:
            scc()

    def __getstate__(self):
        state = super(CacheFSMixin,self).__getstate__()
        state.pop("_CacheFSMixin__cache",None)
        state.pop("_CacheFSMixin__cache_size",None)
        state.pop("_CacheFSMixin__cache_lock",None)
        return state

    def __setstate__(self,state):
        super(CacheFSMixin,self).__setstate__(state)
        self.__cache = PathMap()
        self.__cache_size = 0
        self.__cache_lock = threading.RLock()

    def __get_cached_info(self,path,default=_SENTINAL):
        try:
            info = self.__cache[path]
            if self.cache_timeout is not None:
                now = time.time()
                if info.timestamp < (now - self.cache_timeout):
                    with self.__cache_lock:
                        self.__expire_from_cache(path)
                        raise KeyError
            return info
        except KeyError:
            if default is not _SENTINAL:
                return default
            raise

    def __set_cached_info(self,path,new_ci,old_ci=None):
        was_room = True
        with self.__cache_lock:
            #  Free up some room in the cache
            if self.max_cache_size is not None and old_ci is None:
                while self.__cache_size >= self.max_cache_size:
                    try:
                        to_del = iter(self.__cache).next()
                    except StopIteration:
                        break
                    else:
                        was_room = False
                        self.__expire_from_cache(to_del)
            #  Atomically add to the cache.
            #  If there's a race, newest information wins
            ci = self.__cache.setdefault(path,new_ci)
            if ci is new_ci:
                self.__cache_size += 1
            else:
                if old_ci is None or ci is old_ci:
                    if ci.timestamp < new_ci.timestamp:
                        ci.update_from(new_ci)
        return was_room

    def __expire_from_cache(self,path):
        del self.__cache[path]
        self.__cache_size -= 1
        for ancestor in recursepath(path):
            try:
                self.__cache[ancestor].has_full_children = False
            except KeyError:
                pass

    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        #  Try to validate the entry using the cached info
        try:
            ci = self.__get_cached_info(path)
        except KeyError:
            if path in ("", "/"):
                raise ResourceInvalidError(path)
            try:
                ppath = dirname(path)
                pci = self.__get_cached_info(ppath)
            except KeyError:
                pass
            else:
                if not fs.utils.isdir(super(CacheFSMixin, self), ppath, pci.info):
                    raise ResourceInvalidError(path)
                if pci.has_full_children:
                    raise ResourceNotFoundError(path)
        else:
            if not fs.utils.isfile(super(CacheFSMixin, self), path, ci.info):
                raise ResourceInvalidError(path)
        f = super(CacheFSMixin, self).open(path, mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline, line_buffering=line_buffering, **kwargs)
        if "w" in mode or "a" in mode or "+" in mode:
            with self.__cache_lock:
                self.__cache.clear(path)
            f = self._CacheInvalidatingFile(self, path, f, mode)
        return f

    class _CacheInvalidatingFile(FileWrapper):
        def __init__(self, owner, path, wrapped_file, mode=None):
            self.path = path
            sup = super(CacheFSMixin._CacheInvalidatingFile, self)
            sup.__init__(wrapped_file, mode)
            self.owner = owner
        def _write(self, string, flushing=False):
            with self.owner._CacheFSMixin__cache_lock:
                self.owner._CacheFSMixin__cache.clear(self.path)
            sup = super(CacheFSMixin._CacheInvalidatingFile, self)
            return sup._write(string, flushing=flushing)
        def _truncate(self, size):
            with self.owner._CacheFSMixin__cache_lock:
                self.owner._CacheFSMixin__cache.clear(self.path)
            sup = super(CacheFSMixin._CacheInvalidatingFile, self)
            return sup._truncate(size)

    def exists(self, path):
        try:
            self.getinfo(path)
        except ResourceNotFoundError:
            return False
        else:
            return True

    def isdir(self, path):
        try:
            self.__cache.iternames(path).next()
            return True
        except StopIteration:
            pass
        except RuntimeError:
            pass
        try:
            info = self.getinfo(path)
        except ResourceNotFoundError:
            return False
        else:
            return fs.utils.isdir(super(CacheFSMixin, self), path, info)

    def isfile(self, path):
        try:
            self.__cache.iternames(path).next()
            return False
        except StopIteration:
            pass
        except RuntimeError:
            pass
        try:
            info = self.getinfo(path)
        except ResourceNotFoundError:
            return False
        else:
            return fs.utils.isfile(super(CacheFSMixin, self), path, info)

    def getinfo(self, path):
        try:
            ci = self.__get_cached_info(path)
            if not ci.has_full_info:
                raise KeyError
            info = ci.info
        except KeyError:
            info = super(CacheFSMixin, self).getinfo(path)
            self.__set_cached_info(path, CachedInfo(info))
        return info

    def listdir(self,path="",*args,**kwds):
        return list(nm for (nm, _info) in self.listdirinfo(path,*args,**kwds))

    def ilistdir(self,path="",*args,**kwds):
        for (nm, _info) in self.ilistdirinfo(path,*args,**kwds):
            yield nm

    def listdirinfo(self,path="",*args,**kwds):
        items = super(CacheFSMixin,self).listdirinfo(path,*args,**kwds)
        with self.__cache_lock:
            names = set()
            for (nm,info) in items:
                names.add(basename(nm))
                cpath = pathjoin(path,basename(nm))
                ci = CachedInfo(info)
                self.__set_cached_info(cpath,ci)
            to_del = []
            for nm in self.__cache.names(path):
                if nm not in names:
                    to_del.append(nm)
            for nm in to_del:
                self.__cache.clear(pathjoin(path,nm))
            #try:
            #    pci = self.__cache[path]
            #except KeyError:
            #    pci = CachedInfo.new_dir_stub()
            #    self.__cache[path] = pci
            #pci.has_full_children = True
        return items

    def ilistdirinfo(self,path="",*args,**kwds):
        items = super(CacheFSMixin,self).ilistdirinfo(path,*args,**kwds)
        for (nm,info) in items:
            cpath = pathjoin(path,basename(nm))
            ci = CachedInfo(info)
            self.__set_cached_info(cpath,ci)
            yield (nm,info)

    def getsize(self,path):
        return self.getinfo(path)["size"]

    def setcontents(self, path, data=b'', encoding=None, errors=None, chunk_size=64*1024):
        supsc = super(CacheFSMixin, self).setcontents
        res = supsc(path, data, encoding=None, errors=None, chunk_size=chunk_size)
        with self.__cache_lock:
            self.__cache.clear(path)
            self.__cache[path] = CachedInfo.new_file_stub()
        return res

    def createfile(self, path, wipe=False):
        super(CacheFSMixin,self).createfile(path, wipe=wipe)
        with self.__cache_lock:
            self.__cache.clear(path)
            self.__cache[path] = CachedInfo.new_file_stub()

    def makedir(self,path,*args,**kwds):
        super(CacheFSMixin,self).makedir(path,*args,**kwds)
        with self.__cache_lock:
            self.__cache.clear(path)
            self.__cache[path] = CachedInfo.new_dir_stub()

    def remove(self,path):
        super(CacheFSMixin,self).remove(path)
        with self.__cache_lock:
            self.__cache.clear(path)

    def removedir(self,path,**kwds):
        super(CacheFSMixin,self).removedir(path,**kwds)
        with self.__cache_lock:
            self.__cache.clear(path)

    def rename(self,src,dst):
        super(CacheFSMixin,self).rename(src,dst)
        with self.__cache_lock:
            for (subpath,ci) in self.__cache.items(src):
                self.__cache[pathjoin(dst,subpath)] = ci.clone()
            self.__cache.clear(src)

    def copy(self,src,dst,**kwds):
        super(CacheFSMixin,self).copy(src,dst,**kwds)
        with self.__cache_lock:
            for (subpath,ci) in self.__cache.items(src):
                self.__cache[pathjoin(dst,subpath)] = ci.clone()

    def copydir(self,src,dst,**kwds):
        super(CacheFSMixin,self).copydir(src,dst,**kwds)
        with self.__cache_lock:
            for (subpath,ci) in self.__cache.items(src):
                self.__cache[pathjoin(dst,subpath)] = ci.clone()

    def move(self,src,dst,**kwds):
        super(CacheFSMixin,self).move(src,dst,**kwds)
        with self.__cache_lock:
            for (subpath,ci) in self.__cache.items(src):
                self.__cache[pathjoin(dst,subpath)] = ci.clone()
            self.__cache.clear(src)

    def movedir(self,src,dst,**kwds):
        super(CacheFSMixin,self).movedir(src,dst,**kwds)
        with self.__cache_lock:
            for (subpath,ci) in self.__cache.items(src):
                self.__cache[pathjoin(dst,subpath)] = ci.clone()
            self.__cache.clear(src)

    def settimes(self,path,*args,**kwds):
        super(CacheFSMixin,self).settimes(path,*args,**kwds)
        with self.__cache_lock:
            self.__cache.pop(path,None)


class CacheFS(CacheFSMixin,WrapFS):
    """Simple FS wrapper to cache meta-data of a remote filesystems.

    This FS mixin implements a simplistic cache that can help speed up
    access to a remote filesystem.  File and directory meta-data is cached
    but the actual file contents are not.
    """
    pass





from __future__ import (absolute_import, division, print_function)

import os
import contextlib

import pytest

import numpy as np
from netCDF4 import Dataset

from pyugrid import UGrid
from pyugrid.grid_io import load_from_varnames


@pytest.fixture
@contextlib.contextmanager
def non_compliante_mesh(fname):
    """
    Dummy file based on:
    https://gnome.orr.noaa.gov/py_gnome_testdata/COOPSu_CREOFS.nc

    """
    nc = Dataset(fname, 'w')
    nc.grid_type = 'Triangular'
    nc.createDimension('nbi', 4)
    nc.createDimension('three', 3)
    nc.createDimension('nbnd', 5443)
    nc.createDimension('node', 74061)
    nc.createDimension('nele', 142684)

    bnd = nc.createVariable('bnd', 'i4', dimensions=('nbnd', 'nbi'))
    bnd[:] = np.random.random((5443, 4))

    lon = nc.createVariable('lon', 'f4', dimensions=('node'))
    lon[:] = np.random.random((74061))

    lat = nc.createVariable('lat', 'f4', dimensions=('node'))
    lat[:] = np.random.random((74061))

    nbe = nc.createVariable('nbe', 'i4', dimensions=('three', 'nele'))
    nbe.order = 'ccw'
    nbe[:] = np.random.random((3, 142684))

    nv = nc.createVariable('nv', 'i4', dimensions=('three', 'nele'))
    nv[:] = np.random.random((3, 142684))
    try:
        yield nc
    finally:
        nc.close()
        os.remove(fname)


def test_load_from_varnames_good_mapping():
    mapping = {'attribute_check': ('grid_type', 'triangular'),
               'faces': 'nv',
               'nodes_lon': 'lon',
               'nodes_lat': 'lat',
               'boundaries': 'bnd',
               'face_face_connectivity': 'nbe'}

    fname = 'non_compliant_ugrid.nc'
    with non_compliante_mesh(fname):
        ug = load_from_varnames(fname, mapping)
    assert isinstance(ug, UGrid)


def test_load_from_varnames_bad_mapping():
    mapping = {'attribute_check': ('grid_type', 'triangular'),
               'faces': 'nv',
               'nodes_lon': 'longitude',
               'nodes_lat': 'latitude',
               'boundaries': 'bnd',
               'face_face_connectivity': 'nbe'}

    fname = 'non_compliant_ugrid.nc'
    with non_compliante_mesh(fname):
        with pytest.raises(KeyError):
            load_from_varnames(fname, mapping)



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005, 2006 by Enthought, Inc.
#  All rights reserved.
#
#-----------------------------------------------------------------------------

"""
The single project plugin's project view

"""

# Standard library imports.
import logging
from string import rfind

# Enthought library imports
from apptools.naming.api import Binding
from traits.api import adapts, Any, HasTraits, Instance, Str
from traitsui.api import Item, Group, TreeEditor, ITreeNode, \
    ITreeNodeAdapter, View

# Application specific imports.
from envisage.api import IApplication
from envisage.ui.single_project.project import Project
from envisage.ui.single_project.services import IPROJECT_MODEL, \
    IPROJECT_UI


# Setup a logger for this module.
logger=logging.getLogger(__name__)


# Dummy EmptyProject class for when the ProjectView doesn't have a reference
# to a Project.
class EmptyProject(Project):
    pass

class EmptyProjectAdapter(ITreeNodeAdapter):
    """ Adapter for our EmptyProject. """

    adapts(EmptyProject, ITreeNode)

    #-- ITreeNodeAdapter Method Overrides --------------------------------------

    def get_label(self):
        """ Gets the label to display for a specified object.
        """
        return 'No project loaded.'


class ProjectAdapter(ITreeNodeAdapter):
    """ Base ProjectAdapter for the root of the tree. """

    adapts(Project, ITreeNode)

    #-- ITreeNodeAdapter Method Overrides --------------------------------------

    def allows_children(self):
        """ Returns whether this object can have children.
        """
        return False

    def has_children(self):
        """ Returns whether the object has children.
        """
        return False

    def get_children(self):
        """ Gets the object's children.
        """
        return []

    def get_label(self):
        """ Gets the label to display for a specified object.
        """
        return self.adaptee.name

    def get_tooltip(self):
        """ Gets the tooltip to display for a specified object.
        """
        return "Project"

    def get_icon(self, is_expanded):
        """ Returns the icon for a specified object.
        """
        return '<open>'

    def can_auto_close(self):
        """ Returns whether the object's children should be automatically
            closed.
        """
        return True

class ProjectView(HasTraits):
    """
    The single project plugin's project view

    """

    ##########################################################################
    # Traits
    ##########################################################################

    #### public 'ProjectView' interface ######################################

    # The Envisage application that this service is part of.
    application = Instance(IApplication)

    # The suffix currently applied to our name
    name_suffix = Str('')

    # The suffix applied to titles when the current project is dirty.
    title_suffix = Str('*')

    # Root node for the project.
    root = Instance(Project)

    # The traits view to display:
    view = View(
        Item('root', editor = TreeEditor(editable=False, auto_open=1),
            show_label = False,
        ),
        resizable = True
    )

    ##########################################################################
    # 'View' interface.
    ##########################################################################
    def __init__(self, **traits):
        super(ProjectView, self).__init__(**traits)

        # Make sure our view stays in sync with the current project
        model_service = self._get_model_service()
        model_service.on_trait_change(self._on_project_changed, 'project')
        model_service.on_trait_change(self._on_project_selection_changed,
            'selection')

        # Make sure our control is initialized to the current project.
        self._switch_projects(EmptyProject(), model_service.project)

    ##########################################################################
    # 'ProjectView' interface.
    ##########################################################################

    #### protected 'ProjectView' interface #################################

    def _are_list_contents_different(self, list1, list2):
        """
        Convenience method to determine if two lists contain different
        items.  Returns True if the lists are different and False
        otherwise.

        """
        set1 = set(list1)
        set2 = set(list2)
        return set1 != set2


    def _clear_state(self):
        """
        Clears out all indications of any project state from this view.

        """

        #self.name_suffix = ''
        # Set the root to an EmptyProject.
        self.root = EmptyProject()

        return


    def _get_model_service(self):
        """
        Return a reference to the single project plugin's model service.

        """

        return self.application.get_service(IPROJECT_MODEL)


    def _get_ui_service(self):
        """
        Return a reference to the single project plugin's UI service.

        """

        return self.window.application.get_service(IPROJECT_UI)


    def _switch_projects(self, old, new):
        """
        Switches this view to the specified new project from the old project.

        Either value may be None.

        """

        # Remove listeners on the old project, if any.
        if old is not None and not isinstance(old, EmptyProject):
            self._update_project_listeners(old, remove=True)

        # Update our state according to what the new project is, making sure
        # to add listeners to any new project.
        if new is not None:
            logger.debug("Changing ProjectView to Project [%s]", new.name)
            self._sync_state_to_project(new)
            self._update_project_listeners(new)
        else:
            logger.debug("Changing ProjectView to no project")
            self._clear_state()

        return


    def _sync_state_to_project(self, project):
        """
        Sync the state of this view to the specified project's state.

        """

        logger.debug('Syncing ProjectView [%s] to project state [%s]', self,
            project)

        # Update our Project reference.
        self.root = project

        # Update our name suffix based on the dirty state of the project.
        #self.name_suffix = (project.dirty and self.title_suffix) or ''

        return


    def _update_project_listeners(self, project, remove=False):
        """
        Update listeners on the specified project instance.

        If remove is False then listeners are added, else listeners
        are removed.

        """

        project.on_trait_change(self._on_project_name_changed, 'name', remove)
        project.on_trait_change(self._on_project_dirty_changed, 'dirty',
            remove)

        return


    #### trait handlers ######################################################

    def _name_suffix_changed(self, old, new):
        """
        Handle changes to our name suffix.

        """

        logger.debug('Detected change in name suffix to [%s] within ' + \
            'ProjectView [%s]', new, self)

        # Update our own name by removing the old suffix, if any, and adding
        # on the new suffix, if any.
        name = self.name
        if old is not None and len(old) > 0:
            index = rfind(name, " " + old)
            if index > -1:
                name = name[0:index]
        if new is not None and len(new) > 0:
            name = name + ' ' + new
        self.name = name

        # Update the containing window's suffix as well
        self.window.title_suffix = new

        return


    def _on_control_right_clicked (self, event):
        """
        Handle events when the tree control itself is right-clicked.

        """

        logger.debug('ProjectView control right-clicked')
        self._get_ui_service().display_default_context_menu(self.control, event)

        return


    def _on_key_pressed_changed(self, event):
        """
        Handle key presses while our control has focus.

        """

        logger.debug("ProjectView key pressed [%s]", event)

        # If the delete key was pressed, then delete the current selected
        # object.
        if event.key_code == 127:
            self._get_ui_service().delete_selection()

        return


    def _on_node_activated_changed(self, node):
        """
        Handle nodes being activated (i.e. double-clicked.)

        """

        logger.debug("ProjectView node activated [%s]", node)

        return


    def _on_project_changed(self, obj, trait_name, old, new):
        """
        Handle when the current project changes.

        """

        logger.debug('\n\n ***************** \n\n')
        logger.debug('Detected project changed from [%s] to [%s] in '
            'ProjectView [%s]', old, new, self)

        self._switch_projects(old, new)

        return


    def _on_project_dirty_changed(self, obj, trait_name, old, new):
        """
        Handle the open project's dirty flag changing.

        """

        logger.debug('Detected change in project dirty to [%s] within ' + \
            'ProjectView [%s]', new, self)

        suffix = (new and self.title_suffix) or ''
        #self.name_suffix = suffix

        return


    def _on_project_name_changed(self, obj, trait_name, old, new):
        """
        Handle the open project's name changing.

        """

        self._project_control.root.name = new

        return


    def _on_project_selection_changed(self, obj, trait_name, old, new):
        """
        Handle the current project's selection changing.

        """

        logger.debug('Detected project selection changed from [%s] ' + \
            'to [%s] within ProjectView [%s]', old, new, self)

        # Ensure that the Tree control's selection matches the selection in
        # the project.
        control_selection = self._project_control.selection
        if self._are_list_contents_different(new, control_selection):
            logger.debug('  Updating selection on tree control for ' + \
                'ProjectView [%s]', self)
            self._project_control.set_selection(new)

        # Ensure that this view's selection contains whatever was selected
        # within the project.
        if self._are_list_contents_different(new, self.selection):
            logger.debug('  Updating selection on ProjectView [%s]', self)
            self.selection = new

        return


    def _on_selection_changed(self, obj, trait_name, old, new):
        """
        Handle selection changes in our tree control.

        """

        logger.debug('Detected tree control selection change from [%s] ' + \
            'to [%s] within ProjectView [%s]', old, new, self)

        # Ensure that the project model service's record of selection contains
        # the same elements as the tree control.
        model_service = self._get_model_service()
        if self._are_list_contents_different(new, model_service.selection):
            logger.debug('  Updating selection on project model service')
            model_service.selection = new

        # Note that we don't have to update this view's selection record as
        # we do that through our listener on the project selection.

        return


    def _on_closing_changed(self, old, new):
        """
        Handle when this view closes.

        """

        logger.debug("ProjectView [%s] closing!", self)

        return


#### EOF #####################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   12/19/2004
#
#------------------------------------------------------------------------------

""" Defines the standard menu bar for use with Traits UI windows and panels,
    and standard actions and buttons.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Str

# Import and rename the needed PyFace elements:
from pyface.action.api import ToolBarManager as ToolBar
from pyface.action.api import MenuBarManager as MenuBar
from pyface.action.api import MenuManager as Menu
from pyface.action.api import Group as ActionGroup
from pyface.action.api import Action as PyFaceAction

#-------------------------------------------------------------------------
#  'Action' class (extends the core pyface Action class):
#-------------------------------------------------------------------------


class Action(PyFaceAction):
    """ An action on a menu bar in a Traits UI window or panel.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Pre-condition for showing the action. If the expression evaluates to False,
    # the action is not visible (and disappears if it was previously visible).
    # If the value evaluates to True, the action becomes visible. All
    # **visible_when** conditions are checked each time that any trait value
    # is edited in the display. Therefore, you can use **visible_when**
    # conditions to hide or show actions in response to user input.
    visible_when = Str

    # Pre-condition for enabling the action. If the expression evaluates to
    # False, the action is disabled, that is, it cannot be selected. All
    # **enabled_when** conditions are checked each time that any trait value
    # is edited in the display. Therefore, you can use **enabled_when**
    # conditions to enable or disable actions in response to user input.
    enabled_when = Str

    # Boolean expression indicating when the action is displayed with a check
    # mark beside it. This attribute applies only to actions that are included
    # in menus.
    checked_when = Str

    # Pre-condition for including the action in the menu bar or toolbar. If the
    # expression evaluates to False, the action is not defined in the display.
    # Conditions for **defined_when** are evaluated only once, when the display
    # is first constructed.
    defined_when = Str

    # The method to call to perform the action, on the Handler for the window.
    # The method must accept a single parameter, which is a UIInfo object.
    # Because Actions are associated with Views rather than Handlers, you must
    # ensure that the Handler object for a particular window has a method with
    # the correct name, for each Action defined on the View for that window.
    action = Str

#-------------------------------------------------------------------------
#  Standard actions and menu bar definitions:
#-------------------------------------------------------------------------

# Menu separator:
Separator = ActionGroup

# The standard "close window" action:
CloseAction = Action(
    name='Close',
    action='_on_close'
)

# The standard "undo last change" action:
UndoAction = Action(
    name='Undo',
    action='_on_undo',
    defined_when='ui.history is not None',
    enabled_when='ui.history.can_undo'
)

# The standard "redo last undo" action:
RedoAction = Action(
    name='Redo',
    action='_on_redo',
    defined_when='ui.history is not None',
    enabled_when='ui.history.can_redo'
)

# The standard "revert all changes" action:
RevertAction = Action(
    name='Revert',
    action='_on_revert',
    defined_when='ui.history is not None',
    enabled_when='ui.history.can_undo'
)

# The standard "show help" action:
HelpAction = Action(
    name='Help',
    action='show_help'
)

# The standard Traits UI menu bar:
StandardMenuBar = MenuBar(
    Menu(CloseAction,
         name='File'),
    Menu(UndoAction,
         RedoAction,
         RevertAction,
         name='Edit'),
    Menu(HelpAction,
         name='Help')
)

#-------------------------------------------------------------------------
#  Standard buttons (i.e. actions):
#-------------------------------------------------------------------------

NoButton = Action(name='')

# Appears as two buttons: **Undo** and **Redo**. When **Undo** is clicked, the
# most recent change to the data is cancelled, restoring the previous value.
# **Redo** cancels the most recent "undo" operation.
UndoButton = Action(name='Undo')

# When the user clicks the **Revert** button, all changes made in the window are
# cancelled and the original values are restored. If the changes have been
# applied to the model (because the user clicked **Apply** or because the window
# is live), the model data is restored as well. The window remains open.
RevertButton = Action(name='Revert')

# When theuser clicks the **Apply** button, all changes made in the window are
# applied to the model. This option is meaningful only for modal windows.
ApplyButton = Action(name='Apply')

# When the user clicks the **OK** button, all changes made in the window are
# applied to the model, and the window is closed.
OKButton = Action(name='OK')

# When the user clicks the **Cancel** button, all changes made in the window
# are discarded; if the window is live, the model is restored to the values it
# held before the window was opened. The window is then closed.
CancelButton = Action(name='Cancel')

# When the user clicks the **Help** button, the current help handler is
# invoked. If the default help handler is used, a pop-up window is displayed,
# which contains the **help** text for the top-level Group (if any), and for
# the items in the view. If the default help handler has been overridden,
# the action is determined by the custom help handler. See
# **traitsui.help**.
HelpButton = Action(name='Help')

OKCancelButtons = [OKButton, CancelButton]
ModalButtons = [ApplyButton, RevertButton, OKButton, CancelButton, HelpButton]
LiveButtons = [UndoButton, RevertButton, OKButton, CancelButton, HelpButton]

# The window has no command buttons:
NoButtons = [NoButton]



""" The help code interface.

    This may be used to define examples to be displayed or demos to be run.

    :Copyright: 2008, Enthought Inc.
    :License: BSD
    :Author: Janet Swisher
"""
# This software is provided without warranty under the terms of the BSD
# license included in AppTools/trunk/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!


from traits.api import Interface, File, Str

class IHelpCode(Interface):
    """ The interface for help code.

    A help code is defined by a UI label and a filename.
    """

    # The UI label for the demo, which appears in menus or dialogs.
    label = Str

    # The path to the file containing the code entry point. This may be
    # absolute, or relative to the Python directory (sys.prefix).
    filename = File

    # The unique ID of the preferences node that contains the other values for
    # this object
    preferences_path = Str

    # The code to execute. This is executed when filename is None or an empty
    # string.
    code = Str



# Enthought library imports.
from envisage.ui.tasks.api import PreferencesPane, TaskFactory, TasksApplication
from apptools.preferences.api import PreferencesHelper
from traits.api import on_trait_change, Bool, Dict, Enum, List, Str, Unicode, Instance, Any
from traitsui.api import EnumEditor, HGroup, VGroup, Item, Label, \
    View
from envisage.api import IApplication
from pyface.api import Dialog, OK


class FrameworkPreferenceDialog(Dialog):

    application = Instance(IApplication)

    def _create_dialog_area(self, parent):
        from envisage.ui.tasks.preferences_dialog import \
            PreferencesDialog

        dialog = self.application.get_service(PreferencesDialog)
        self.prefs_ui = dialog.edit_traits(parent=parent, scrollable=True, kind='subpanel')
        return self.prefs_ui.control

    @on_trait_change('closed')
    def delete_prefs_ui(self):
        if self.return_code == OK:
            self.prefs_ui.handler.apply()
            self.application.preferences.save()
        self.prefs_ui.finish()
        self.prefs_ui = None


class FrameworkPreferences(PreferencesHelper):
    """ The preferences helper for the Framework application.
    """

    #### 'PreferencesHelper' interface ########################################

    # Enthought bug? The application trait is needed to provide subclasses a
    # way to find application level defaults, like user config directories.
    # But, if I declare this to be a trait, upon assignment in
    # FrameworkApplication.get_preferences, the FrameworkApplication instance
    # gets converted to unicode.
    #
    # application = Any

    # Instead, use a class attribute because the application object won't
    # change so there's no need to make it an instance attribute.
    application = None

    # The path to the preference node that contains the preferences.
    preferences_path = 'omnivore.framework'

    #### Preferences ##########################################################

    # The task to activate on app startup if not restoring an old layout.
    default_task = Str

    # Whether to always apply the default application-level layout.
    # See TasksApplication for more information.
    always_use_default_layout = Bool


class FrameworkPreferencesPane(PreferencesPane):
    """ The preferences pane for the Framework application.
    """

    #### 'PreferencesPane' interface ##########################################

    # The factory to use for creating the preferences model object.
    model_factory = FrameworkPreferences

    #### 'FrameworkPreferencesPane' interface ################################

    task_map = Dict(Str, Unicode)

    view = View(
        VGroup(HGroup(Item('always_use_default_layout'),
                      Label('Always use the default active task on startup'),
                      show_labels = False),
               HGroup(Label('Default active task:'),
                      Item('default_task',
                           editor=EnumEditor(name='handler.task_map')),
                      enabled_when = 'always_use_default_layout',
                      show_labels = False),
               label='Application startup'),
        resizable=True)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _task_map_default(self):
        return dict((factory.id, factory.name)
                    for factory in self.dialog.application.task_factories)



""" An action that resets the active perspective. """


# Enthought library imports.
from pyface.api import YES

# Local imports.
from .workbench_action import WorkbenchAction


# The message used when confirming the action.
MESSAGE = 'Do you want to reset the current "%s" perspective to its defaults?'


class ResetActivePerspectiveAction(WorkbenchAction):
    """ An action that resets the active perspective. """

    #### 'Action' interface ###################################################

    # The action's unique identifier (may be None).
    id = 'pyface.workbench.action.reset_active_perspective'

    # The action's name (displayed on menus/tool bar tools etc).
    name = 'Reset Perspective'

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Perform the action. """

        window = self.window

        if window.confirm(MESSAGE % window.active_perspective.name) == YES:
            window.reset_active_perspective()

        return

#### EOF ######################################################################



import os, time, logging, threading, multiprocessing, Queue

# Utilities for thread and process based jobs


# Don't use the logger module in anything using multiprocessing! It can cause
# deadlocks.  I.e. don't use this:
#### log = logging.getLogger("dinoteeth.job")

log = multiprocessing.log_to_stderr()


class Job(object):
    def __init__(self, job_id=None):
        self.job_id = job_id
        self.parent = None
        self.children_running = 0
        self.children_scheduled = []
        self.error = None
        self.exception = None

    def debug(self, s):
        log.debug(s)

    def __str__(self):
        status = self.get_name()
        if not self.success():
            if self.error is not None:
                status += " Failed: %s" % self.error
            else:
                status += " Exception: %s" % self.exception
        return status

    def success(self):
        return self.error is None and self.exception is None

    def get_name(self):
        return self.__class__.__name__

    def _start(self, dispatcher):
        raise RuntimeError("Abstract method")

    def _success_message(self):
        return None

    def _failed_message(self):
        return None

    def _get_status_message(self):
        if self.success():
            return self._success_message()
        else:
            return self._failed_message()

    def success_callback(self):
        """Called in main thread if job completes successfully.
        
        Note that since it occurs in the main thread, GUI methods may
        be safely called.
        """
        pass

    def failure_callback(self):
        """Called in main thread if job fails during the thread processing.
        
        Note that since it occurs in the main thread, GUI methods may be safely
        called.
        """
        pass


class ProcessJob(Job):
    def _start(self, results):
        raise RuntimeError("Abstract method")


class LargeMemoryJob(Job):
    def _start(self, results):
        raise RuntimeError("Abstract method")


class Worker(multiprocessing.Process):
    def __init__(self, job_queue, progress_queue):
        multiprocessing.Process.__init__(self)
        self._jobs = job_queue
        self._progress = progress_queue
        self.start()

    def _progress_update(self, item):
        update = ProgressReport(item)
        self._progress.put(update)

    def run(self):
        while True:
            log.debug("%s: worker waiting for job" % self.name)
            job = self._jobs.get() # block to wait for new job
            log.debug("%s: worker found job %s" % (self.name, str(job)))
            if job is None:
                # "poison pill" means shutdown this worker
                log.debug("%s: poison pill received. Stopping" % self.name)
                self._progress.put(Shutdown())
                break
            try:
                job._start(self)
            except Exception, e:
                import traceback
                job.exception = traceback.format_exc()
            self._progress.put(Finished(job))


class JobDispatcher(object):
    def __init__(self, share_input_queue_with=None):
        self._want_abort = False
        if share_input_queue_with is not None:
            self._queue = share_input_queue_with._queue
        else:
            self._queue = Queue.Queue()

    def set_manager(self, manager):
        self._manager = manager

    def start_processing(self):
        raise RuntimeError("Abstract method")

    @classmethod
    def can_handle(self, job):
        return False

    def add_job(self, job):
        self._queue.put(job)

    def abort(self):
        # Method for use by main thread to signal an abort
        self._want_abort = True
        self._queue.put(None)


class ThreadJobDispatcher(threading.Thread, JobDispatcher):
    def __init__(self, share_input_queue_with=None):
        threading.Thread.__init__(self)
        JobDispatcher.__init__(self, share_input_queue_with)
        self.log = log

    def start_processing(self):
#        self.setDaemon(True)
        self.start()

    @classmethod
    def can_handle(self, job):
        return isinstance(job, ThreadJob)

    def run(self):
        log.debug("starting thread...")
        while True:
            log.debug("%s waiting for jobs..." % self.name)
            job = self._queue.get(True) # blocking
            if job is None or self._want_abort:
                break
            try:
                job._start(self)
            except Exception, e:
                import traceback
                job.exception = traceback.format_exc()
            self._manager._job_done(job)


class ProcessJobDispatcher(ThreadJobDispatcher):
    def __init__(self, *args, **kwargs):
        ThreadJobDispatcher.__init__(self, *args, **kwargs)
        self._multiprocessing_jobs = multiprocessing.Queue()
        self._multiprocessing_progress = multiprocessing.Queue()
        self._worker = Worker(self._multiprocessing_jobs, self._multiprocessing_progress)
        log.debug("worker %s: status = %s" % (self._worker, self._worker.exitcode))

    @classmethod
    def can_handle(self, job):
        return isinstance(job, ProcessJob)

    def run(self):
        log.debug("%s: starting process job dispatcher thread..." % self.name)
        while True:
            log.debug("%s: waiting for jobs..." % self.name)
            job = self._queue.get(True) # blocking

            # Send job to worker and wait for it to finish
            log.debug("%s: sending job '%s' to process %s" % (self.name, job, self._worker))
            self._multiprocessing_jobs.put(job)
            while True:
                progress = self._multiprocessing_progress.get(True)
                if isinstance(progress, Finished):
                    self._manager._job_done(progress.job)
                    break
                elif isinstance(progress, Shutdown):
                    job = None
                    break
                else:
                    self._manager._progress_report(progress.report)

            if job is None:
                log.debug("%s: Stopping process %s" % (self.name, self._worker))
                self._worker.join()
                log.debug("%s: Exiting dispatcher %s" % (self.name, self.name))
                break


class ProgressReport(object):
    def __init__(self, job_id=None, report=None):
        self.job_id = job_id
        self.report = report

    def is_finished(self):
        return False


class Shutdown(ProgressReport):
    pass


class Finished(ProgressReport):
    def is_finished(self):
        return True


class Running(ProgressReport):
    pass


class Terminated(ProgressReport):
    pass


class LargeMemoryWorker(multiprocessing.Process):
    def __init__(self, job, progress_queue):
        multiprocessing.Process.__init__(self)
        self._job = job
        self._progress = progress_queue

    def _progress_update(self, item):
        self._progress.put(item)

    def run(self):
        self._progress.put(Running())
        try:
            self._job._start(self)
        except Exception, e:
            import traceback
            self._job.exception = traceback.format_exc()
        self._progress.put(None)


class LargeMemoryJobDispatcher(ThreadJobDispatcher):
    def __init__(self, *args, **kwargs):
        ThreadJobDispatcher.__init__(self, *args, **kwargs)
        self._multiprocessing_progress = multiprocessing.Queue()
        self._worker = None
        self._is_running = False
        self._timeout = 5

    @classmethod
    def can_handle(self, job):
        return isinstance(job, LargeMemoryJob)

    def add_job(self, job):
        self._worker = LargeMemoryWorker(job, self._multiprocessing_progress)
        log.debug("LARGEMEM: worker %s: status = %s" % (self._worker, self._worker.exitcode))
        self.start()

    def run(self):
        log.debug("LARGEMEM: %s: starting %s..." % (self.name, self._worker))
        self._worker.start()
        while True:
            log.debug("LARGEMEM: %s: waiting for progress queue" % (self.name))
            try:
                progress = self._multiprocessing_progress.get(True, timeout=1)
                log.debug("LARGEMEM: %s: got progress update from queue" % (self.name))
                if progress is None:
                    break
                elif isinstance(progress, Running):
                    self._is_running = True
                else:
                    self._manager._progress_report(progress)
            except Queue.Empty:
                log.debug("LARGEMEM: %s: progress queue empty; worker %s status=%s alive=%s" % (self.name, self._worker, self._worker.exitcode, self._worker.is_alive()))
            if not self._is_running:
                log.debug("LARGEMEM: %s: waiting for confirmation that worker is running. %d seconds remaining" % (self.name, self._timeout))
                self._timeout -= 1
                if self._timeout <= 0:
                    log.debug("LARGEMEM: %s: didn't get confirmation that worker is running. Force quit!" % (self.name))
                    self._manager._progress_report(Terminated())
                    self._worker.terminate()
                    break
        log.debug("LARGEMEM: %s: Stopping process %s" % (self.name, self._worker))
        self._worker.join()
        log.debug("LARGEMEM: %s: Exiting dispatcher %s" % (self.name, self.name))
        self._manager._job_done(None, self)


class Timer(threading.Thread):
    def __init__(self, event_callback, resolution=.2):
        threading.Thread.__init__(self)
        self._event_callback = event_callback
        self._event = threading.Event()
        self._resolution = resolution
        self._expire_time = time.time()
        self._want_abort = False
        self.start()

    def run(self):
        while self._event.wait():
            while self._event.is_set():
                if self._want_abort:
                    return
                time.sleep(self._resolution)
#                log.debug("timer!!!!")
                self._event_callback()
                if time.time() > self._expire_time:
                    self.stop_ticks()
            if self._want_abort:
                return
            self._event_callback()

    def start_ticks(self, resolution, expire_time):
        self._resolution = resolution
        self._expire_time = expire_time
        self._event.set()

    def stop_ticks(self):
        if not self._want_abort:
            self._event.clear()
        log.debug("Timer.stop_ticks: timer expired!!!")

    def abort(self):
        log.debug("Timer.abort: stopping timer!!!")
        self._want_abort = True
        self._event.set()


class JobManager(object):
    def __init__(self, event_callback):
        log = logging.getLogger(self.__class__.__name__)
        self.event_callback = event_callback
        self.job_id_handlers = {}
        self._finished = Queue.Queue()
        self.dispatchers = []
        self.dispatcher_classes = [LargeMemoryJobDispatcher]
        self.timer = Timer(event_callback)

    def start_ticks(self, resolution, expire_time):
        self.timer.start_ticks(resolution, expire_time)

    def stop_ticks(self):
        self.timer.stop_ticks()

    def find_dispatcher(self, job):
        for dispatcher in self.dispatchers:
            if dispatcher.can_handle(job):
                return dispatcher
        for dispatcher_cls in self.dispatcher_classes:
            if dispatcher_cls.can_handle(job):
                dispatcher = dispatcher_cls()
                dispatcher.set_manager(self)
                return dispatcher
        return None

    def start_dispatcher(self, dispatcher):
        log.debug("Adding dispatcher %s" % str(dispatcher))
        dispatcher.set_manager(self)
        dispatcher.start_processing()
        self.dispatchers.append(dispatcher)

    def add_job(self, job):
        dispatcher = self.find_dispatcher(job)
        if dispatcher is not None:
            log.debug("Adding job %s to %s" % (str(job), str(dispatcher)))
            dispatcher.add_job(job)
        else:
            log.debug("No dispatcher for job %s" % str(job))
        return dispatcher is not None

    def _progress_report(self, progress_report):
        """Called from threads to report milestones as the job works
        
        If event handler is defined, also calls that function to report to the
        UI that a job is available.
        """
        log.debug("progress report in thread: %s" % repr(progress_report))
        if self.event_callback is not None:
            self.event_callback(progress_report)

    def _job_done(self, job, dispatcher=None):
        """Called from threads to report completed jobs
        
        If event handler is defined, also calls that function to report to the
        UI that a job is available.
        """
        if job is not None:
            self._finished.put(job)
        if dispatcher is not None:
            self._finished.put(dispatcher)
        if self.event_callback is not None:
            if job is not None:
                message = job._get_status_message()
            else:
                message = None
            self.event_callback(message)

    def get_finished(self):
        done = set()
        try:
            while True:
                item = self._finished.get(False)
                if hasattr(item, "can_handle"):
                    # really a dispatcher
                    log.debug("dispatcher %s completed" % str(item))
                    item.join()
                    log.debug("dispatcher %s joined" % str(item))
                else:
                    done.add(item)
        except Queue.Empty:
            pass
        for job in done:
            log.debug("job %s completed" % str(job))
            if job is None:
                # Skip any poison pill jobs
                continue
            if job.parent is not None:
                log.debug("  subjob of %s" % str(job.parent))
            if job.success():
                job.success_callback()
            else:
                job.failure_callback()
        return done

    def shutdown(self):
        for dispatcher in self.dispatchers:
            dispatcher.abort()
        self.timer.abort()
        for dispatcher in self.dispatchers:
            dispatcher.join()
        self.timer.join()

    def register_job_id_callback(self, job_id, callback):
        self.job_id_handlers[job_id] = callback

    def handle_job_id_callback(self, event):
        if hasattr(event, 'job_id'):
            job_id = event.job_id
            callback = self.job_id_handlers.get(job_id, None)
            if callback is not None:
                log.debug("handle_job_id_callback: found callback for %s!" % job_id)
                callback(event)
                if isinstance(event, Finished):
                    # automatically remove handler
                    del self.job_id_handlers[job_id]
            else:
                log.debug("handle_job_id_callback: no callback for %s!" % job_id)
        else:
            log.debug("handle_job_id_callback: no callback for generic event %s!" % event)


GlobalJobManager = None


def create_global_job_manager(callback):
    global GlobalJobManager
    if GlobalJobManager is None:
        GlobalJobManager = JobManager(callback)
    return GlobalJobManager


def get_global_job_manager():
    global GlobalJobManager
    return GlobalJobManager


if __name__ == '__main__':
    import functools

    class TestProcessSleepJob(ProcessJob):
        def __init__(self, num, sleep):
            Job.__init__(self)
            self.num = num
            self.sleep = sleep

        def get_name(self):
            return "process sleep job #%d, delay=%ss" % (self.num, self.sleep)

        def _start(self, dispatcher):
            log.debug("%s starting!" % self.get_name())
            time.sleep(self.sleep)
            dispatcher._progress_update("Update #1 for %d" % self.num)
            time.sleep(self.sleep)
            dispatcher._progress_update("Update #2 for %d" % self.num)
            time.sleep(self.sleep)
            dispatcher._progress_update("Update #3 for %d" % self.num)

    def post_event(event_name, *args):
        print "event: %s.  args=%s" % (event_name, str(args))

    def get_event_callback(event):
        callback = functools.partial(post_event, event)
        return callback

    def test_sleep():
        callback = get_event_callback("on_status_change")
        manager = JobManager(callback)
        dispatcher3 = ProcessJobDispatcher()
        manager.start_dispatcher(dispatcher3)
        for i in range(3):
            dispatcher_i = ProcessJobDispatcher(dispatcher3)
            manager.start_dispatcher(dispatcher_i)

        manager.add_job(TestProcessSleepJob(10, .1))
        manager.add_job(TestProcessSleepJob(11, .1))
        manager.add_job(TestProcessSleepJob(12, .3))
        manager.add_job(TestProcessSleepJob(13, .1))
        manager.add_job(TestProcessSleepJob(14, .1))
        for i in range(5):
            time.sleep(1)
            jobs = manager.get_finished()
            for job in jobs:
                print 'FINISHED:', str(job)

    #    manager.add_job(TestProcessSleepJob(6, 1))
        manager.shutdown()
        jobs = manager.get_finished()
        for job in jobs:
            print 'FINISHED:', str(job)
        for i in range(5):
            time.sleep(1)
            jobs = manager.get_finished()
            for job in jobs:
                print 'FINISHED:', str(job)

    test_sleep()



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#
#------------------------------------------------------------------------------

""" Defines the editor factory for multi-selection enumerations, for all traits toolkit
backends.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Range

from ..editor_factory import EditorWithListFactory

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorWithListFactory):
    """ Editor factory for checklists.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Number of columns to use when the editor is displayed as a grid
    cols = Range(1, 20)

# Define the CheckListEditor class
CheckListEditor = ToolkitEditorFactory



"""
fs.path
=======

Useful functions for FS path manipulation.

This is broadly similar to the standard ``os.path`` module but works with
paths in the canonical format expected by all FS objects (that is, separated
by forward slashes and with an optional leading slash).

"""

import re
import os


_requires_normalization = re.compile(r'(^|/)\.\.?($|/)|//').search


def normpath(path):
    """Normalizes a path to be in the format expected by FS objects.

    This function removes trailing slashes, collapses duplicate slashes,
    and generally tries very hard to return a new path in the canonical FS format.
    If the path is invalid, ValueError will be raised.

    :param path: path to normalize
    :returns: a valid FS path

    >>> normpath("/foo//bar/frob/../baz")
    '/foo/bar/baz'

    >>> normpath("foo/../../bar")
    Traceback (most recent call last)
        ...
    BackReferenceError: Too many backrefs in 'foo/../../bar'

    """

    if path in ('', '/'):
        return path

    # An early out if there is no need to normalize this path
    if not _requires_normalization(path):
        return path.rstrip('/')

    prefix = u'/' if path.startswith('/') else u''
    components = []
    append = components.append
    special = ('..', '.', '').__contains__
    try:
        for component in path.split('/'):
            if special(component):
                if component == '..':
                    components.pop()
            else:
                append(component)
    except IndexError:
        # Imported here because errors imports this module (path),
        # causing a circular import.
        from fs.errors import BackReferenceError
        raise BackReferenceError('Too many backrefs in \'%s\'' % path)
    return prefix + u'/'.join(components)


if os.sep != '/':
    def ospath(path):
        """Replace path separators in an OS path if required"""
        return path.replace(os.sep, '/')
else:
    def ospath(path):
        """Replace path separators in an OS path if required"""
        return path


def iteratepath(path, numsplits=None):
    """Iterate over the individual components of a path.

    :param path: Path to iterate over
    :numsplits: Maximum number of splits

    """
    path = relpath(normpath(path))
    if not path:
        return []
    if numsplits == None:
        return path.split('/')
    else:
        return path.split('/', numsplits)


def recursepath(path, reverse=False):
    """Returns intermediate paths from the root to the given path

    :param reverse: reverses the order of the paths

    >>> recursepath('a/b/c')
    ['/', u'/a', u'/a/b', u'/a/b/c']

    """

    if path in ('', '/'):
        return [u'/']

    path = abspath(normpath(path)) + '/'

    paths = [u'/']
    find = path.find
    append = paths.append
    pos = 1
    len_path = len(path)

    while pos < len_path:
        pos = find('/', pos)
        append(path[:pos])
        pos += 1

    if reverse:
        return paths[::-1]
    return paths


def isabs(path):
    """Return True if path is an absolute path."""
    return path.startswith('/')


def abspath(path):
    """Convert the given path to an absolute path.

    Since FS objects have no concept of a 'current directory' this simply
    adds a leading '/' character if the path doesn't already have one.

    """
    if not path.startswith('/'):
        return u'/' + path
    return path


def relpath(path):
    """Convert the given path to a relative path.

    This is the inverse of abspath(), stripping a leading '/' from the
    path if it is present.

    :param path: Path to adjust

    >>> relpath('/a/b')
    'a/b'

    """
    return path.lstrip('/')


def pathjoin(*paths):
    """Joins any number of paths together, returning a new path string.

    :param paths: Paths to join are given in positional arguments

    >>> pathjoin('foo', 'bar', 'baz')
    'foo/bar/baz'

    >>> pathjoin('foo/bar', '../baz')
    'foo/baz'

    >>> pathjoin('foo/bar', '/baz')
    '/baz'

    """
    absolute = False
    relpaths = []
    for p in paths:
        if p:
            if p[0] == '/':
                del relpaths[:]
                absolute = True
            relpaths.append(p)

    path = normpath(u"/".join(relpaths))
    if absolute:
        path = abspath(path)
    return path


def pathcombine(path1, path2):
    """Joins two paths together.

    This is faster than `pathjoin`, but only works when the second path is relative,
    and there are no backreferences in either path.

    >>> pathcombine("foo/bar", "baz")
    'foo/bar/baz'

    """
    if not path1:
        return path2.lstrip()
    return "%s/%s" % (path1.rstrip('/'), path2.lstrip('/'))


def join(*paths):
    """Joins any number of paths together, returning a new path string.

    This is a simple alias for the ``pathjoin`` function, allowing it to be
    used as ``fs.path.join`` in direct correspondence with ``os.path.join``.

    :param paths: Paths to join are given in positional arguments
    """
    return pathjoin(*paths)


def pathsplit(path):
    """Splits a path into (head, tail) pair.

    This function splits a path into a pair (head, tail) where 'tail' is the
    last pathname component and 'head' is all preceding components.

    :param path: Path to split

    >>> pathsplit("foo/bar")
    ('foo', 'bar')

    >>> pathsplit("foo/bar/baz")
    ('foo/bar', 'baz')

    >>> pathsplit("/foo/bar/baz")
    ('/foo/bar', 'baz')

    """
    if '/' not in path:
        return ('', path)
    split = path.rsplit('/', 1)
    return (split[0] or '/', split[1])


def split(path):
    """Splits a path into (head, tail) pair.

    This is a simple alias for the ``pathsplit`` function, allowing it to be
    used as ``fs.path.split`` in direct correspondence with ``os.path.split``.

    :param path: Path to split
    """
    return pathsplit(path)


def splitext(path):
    """Splits the extension from the path, and returns the path (up to the last
    '.' and the extension).

    :param path: A path to split

    >>> splitext('baz.txt')
    ('baz', 'txt')

    >>> splitext('foo/bar/baz.txt')
    ('foo/bar/baz', 'txt')

    """

    parent_path, pathname = pathsplit(path)
    if '.' not in pathname:
        return path, ''
    pathname, ext = pathname.rsplit('.', 1)
    path = pathjoin(parent_path, pathname)
    return path, '.' + ext


def isdotfile(path):
    """Detects if a path references a dot file, i.e. a resource who's name
    starts with a '.'

    :param path: Path to check

    >>> isdotfile('.baz')
    True

    >>> isdotfile('foo/bar/.baz')
    True

    >>> isdotfile('foo/bar.baz')
    False

    """
    return basename(path).startswith('.')


def dirname(path):
    """Returns the parent directory of a path.

    This is always equivalent to the 'head' component of the value returned
    by pathsplit(path).

    :param path: A FS path

    >>> dirname('foo/bar/baz')
    'foo/bar'

    >>> dirname('/foo/bar')
    '/foo'

    >>> dirname('/foo')
    '/'

    """
    return pathsplit(path)[0]


def basename(path):
    """Returns the basename of the resource referenced by a path.

    This is always equivalent to the 'tail' component of the value returned
    by pathsplit(path).

    :param path: A FS path

    >>> basename('foo/bar/baz')
    'baz'

    >>> basename('foo/bar')
    'bar'

    >>> basename('foo/bar/')
    ''

    """
    return pathsplit(path)[1]


def issamedir(path1, path2):
    """Return true if two paths reference a resource in the same directory.

    :param path1: An FS path
    :param path2: An FS path

    >>> issamedir("foo/bar/baz.txt", "foo/bar/spam.txt")
    True
    >>> issamedir("foo/bar/baz/txt", "spam/eggs/spam.txt")
    False

    """
    return dirname(normpath(path1)) == dirname(normpath(path2))


def isbase(path1, path2):
    p1 = forcedir(abspath(path1))
    p2 = forcedir(abspath(path2))
    return p1 == p2 or p1.startswith(p2)


def isprefix(path1, path2):
    """Return true is path1 is a prefix of path2.

    :param path1: An FS path
    :param path2: An FS path

    >>> isprefix("foo/bar", "foo/bar/spam.txt")
    True
    >>> isprefix("foo/bar/", "foo/bar")
    True
    >>> isprefix("foo/barry", "foo/baz/bar")
    False
    >>> isprefix("foo/bar/baz/", "foo/baz/bar")
    False

    """
    bits1 = path1.split("/")
    bits2 = path2.split("/")
    while bits1 and bits1[-1] == "":
        bits1.pop()
    if len(bits1) > len(bits2):
        return False
    for (bit1, bit2) in zip(bits1, bits2):
        if bit1 != bit2:
            return False
    return True


def forcedir(path):
    """Ensure the path ends with a trailing forward slash

    :param path: An FS path

    >>> forcedir("foo/bar")
    'foo/bar/'
    >>> forcedir("foo/bar/")
    'foo/bar/'

    """

    if not path.endswith('/'):
        return path + '/'
    return path


def frombase(path1, path2):
    if not isprefix(path1, path2):
        raise ValueError("path1 must be a prefix of path2")
    return path2[len(path1):]


def relativefrom(base, path):
    """Return a path relative from a given base path,
    i.e. insert backrefs as appropriate to reach the path from the base.

    :param base_path: Path to a directory
    :param path: Path you wish to make relative


    >>> relativefrom("foo/bar", "baz/index.html")
    '../../baz/index.html'

    """
    base = list(iteratepath(base))
    path = list(iteratepath(path))

    common = 0
    for a, b in zip(base, path):
        if a != b:
            break
        common += 1

    return u'/'.join([u'..'] * (len(base) - common) + path[common:])


class PathMap(object):
    """Dict-like object with paths for keys.

    A PathMap is like a dictionary where the keys are all FS paths.  It has
    two main advantages over a standard dictionary.  First, keys are normalized
    automatically::

        >>> pm = PathMap()
        >>> pm["hello/world"] = 42
        >>> print pm["/hello/there/../world"]
        42

    Second, various dictionary operations (e.g. listing or clearing values)
    can be efficiently performed on a subset of keys sharing some common
    prefix::

        # list all values in the map
        pm.values()

        # list all values for paths starting with "/foo/bar"
        pm.values("/foo/bar")

    Under the hood, a PathMap is a trie-like structure where each level is
    indexed by path name component.  This allows lookups to be performed in
    O(number of path components) while permitting efficient prefix-based
    operations.
    """

    def __init__(self):
        self._map = {}

    def __getitem__(self, path):
        """Get the value stored under the given path."""
        m = self._map
        for name in iteratepath(path):
            try:
                m = m[name]
            except KeyError:
                raise KeyError(path)
        try:
            return m[""]
        except KeyError:
            raise KeyError(path)

    def __contains__(self, path):
        """Check whether the given path has a value stored in the map."""
        try:
            self[path]
        except KeyError:
            return False
        else:
            return True

    def __setitem__(self, path, value):
        """Set the value stored under the given path."""
        m = self._map
        for name in iteratepath(path):
            try:
                m = m[name]
            except KeyError:
                m = m.setdefault(name, {})
        m[""] = value

    def __delitem__(self, path):
        """Delete the value stored under the given path."""
        ms = [[self._map, None]]
        for name in iteratepath(path):
            try:
                ms.append([ms[-1][0][name], None])
            except KeyError:
                raise KeyError(path)
            else:
                ms[-2][1] = name
        try:
            del ms[-1][0][""]
        except KeyError:
            raise KeyError(path)
        else:
            while len(ms) > 1 and not ms[-1][0]:
                del ms[-1]
                del ms[-1][0][ms[-1][1]]

    def get(self, path, default=None):
        """Get the value stored under the given path, or the given default."""
        try:
            return self[path]
        except KeyError:
            return default

    def pop(self, path, default=None):
        """Pop the value stored under the given path, or the given default."""
        ms = [[self._map, None]]
        for name in iteratepath(path):
            try:
                ms.append([ms[-1][0][name], None])
            except KeyError:
                return default
            else:
                ms[-2][1] = name
        try:
            val = ms[-1][0].pop("")
        except KeyError:
            val = default
        else:
            while len(ms) > 1 and not ms[-1][0]:
                del ms[-1]
                del ms[-1][0][ms[-1][1]]
        return val

    def setdefault(self, path, value):
        m = self._map
        for name in iteratepath(path):
            try:
                m = m[name]
            except KeyError:
                m = m.setdefault(name, {})
        return m.setdefault("", value)

    def clear(self, root="/"):
        """Clear all entries beginning with the given root path."""
        m = self._map
        for name in iteratepath(root):
            try:
                m = m[name]
            except KeyError:
                return
        m.clear()

    def iterkeys(self, root="/", m=None):
        """Iterate over all keys beginning with the given root path."""
        if m is None:
            m = self._map
            for name in iteratepath(root):
                try:
                    m = m[name]
                except KeyError:
                    return
        for (nm, subm) in m.iteritems():
            if not nm:
                yield abspath(root)
            else:
                k = pathcombine(root, nm)
                for subk in self.iterkeys(k, subm):
                    yield subk

    def __iter__(self):
        return self.iterkeys()

    def keys(self,root="/"):
        return list(self.iterkeys(root))

    def itervalues(self, root="/", m=None):
        """Iterate over all values whose keys begin with the given root path."""
        root = normpath(root)
        if m is None:
            m = self._map
            for name in iteratepath(root):
                try:
                    m = m[name]
                except KeyError:
                    return
        for (nm, subm) in m.iteritems():
            if not nm:
                yield subm
            else:
                k = pathcombine(root, nm)
                for subv in self.itervalues(k, subm):
                    yield subv

    def values(self, root="/"):
        return list(self.itervalues(root))

    def iteritems(self, root="/", m=None):
        """Iterate over all (key,value) pairs beginning with the given root."""
        root = normpath(root)
        if m is None:
            m = self._map
            for name in iteratepath(root):
                try:
                    m = m[name]
                except KeyError:
                    return
        for (nm, subm) in m.iteritems():
            if not nm:
                yield (abspath(normpath(root)), subm)
            else:
                k = pathcombine(root, nm)
                for (subk, subv) in self.iteritems(k, subm):
                    yield (subk, subv)

    def items(self, root="/"):
        return list(self.iteritems(root))

    def iternames(self, root="/"):
        """Iterate over all names beneath the given root path.

        This is basically the equivalent of listdir() for a PathMap - it yields
        the next level of name components beneath the given path.
        """
        m = self._map
        for name in iteratepath(root):
            try:
                m = m[name]
            except KeyError:
                return
        for (nm, subm) in m.iteritems():
            if nm and subm:
                yield nm

    def names(self, root="/"):
        return list(self.iternames(root))


_wild_chars = frozenset('*?[]!{}')


def iswildcard(path):
    """Check if a path ends with a wildcard

    >>> is_wildcard('foo/bar/baz.*')
    True
    >>> is_wildcard('foo/bar')
    False

    """
    assert path is not None
    return not _wild_chars.isdisjoint(path)

if __name__ == "__main__":
    print recursepath('a/b/c')

    print relativefrom('/', '/foo')
    print relativefrom('/foo/bar', '/foo/baz')
    print relativefrom('/foo/bar/baz', '/foo/egg')
    print relativefrom('/foo/bar/baz/egg', '/foo/egg')



# Standard library imports.
import sys

# Major package imports.
import wx
from wx.lib.ClickableHtmlWindow import PyClickableHtmlWindow

# Enthought library imports.
from traits.api import Bool, Event, Instance, File, Unicode, Property, provides

# Local imports.
from omnivore.framework.editor import FrameworkEditor


class HtmlWindow(PyClickableHtmlWindow):
    def OnCellMouseHover(self, cell, x, y):
        # Without access to the task window, search the control hierarchy to
        # find a wx.Frame and set the status text directly
        parent = self.GetParent()
        while parent:
            if hasattr(parent, "SetStatusText"):
                linkinfo = cell.GetLink()
                if linkinfo is not None:
                    parent.SetStatusText(linkinfo.GetHref())
                else:
                    parent.SetStatusText("")
                return
            parent = parent.GetParent()


class HtmlViewer(FrameworkEditor):
    """ The toolkit specific implementation of a simple HTML viewer
    """

    #### Events ####

    changed = Event

    ###########################################################################
    # 'PythonEditor' interface.
    ###########################################################################

    def create(self, parent):
        self.control = self._create_control(parent)

    def rebuild_document_properties(self):
        text = self.document.to_bytes()
        self.control.SetPage(text)

    def set_preferences(self):
        prefs = self.task.preferences
        self.control.SetStandardFonts(prefs.font_size, prefs.normal_face, prefs.fixed_face)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_control(self, parent):
        """ Creates the toolkit-specific control for the widget. """

        # Base-class constructor.
        self.control = HtmlWindow(parent, -1, style=wx.NO_FULL_REPAINT_ON_RESIZE)
        self.set_preferences()

        # Load the editor's contents.
        self.load()

        return self.control



#------------------------------------------------------------------------------
# Copyright (c) 2008, Enthought, Inc.
# All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#
#------------------------------------------------------------------------------
""" Defines the file editor factory for all traits toolkit backends.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import List, Str, Bool, Int, Unicode, File

# CIRCULAR IMPORT FIXME: Importing from the source rather than traits.ui.api
# to avoid circular imports, as this EditorFactory will be part of
# traits.ui.api as well.
from ..view import View

from ..group import Group

from .text_editor import ToolkitEditorFactory as EditorFactory

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Wildcard filter:
filter_trait = List(Unicode)

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for file editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Wildcard filter to apply to the file dialog:
    filter = filter_trait

    # Optional extended trait name of the trait containing the list of filters:
    filter_name = Str

    # Should file extension be truncated?
    truncate_ext = Bool(False)

    # Can the user select directories as well as files?
    allow_dir = Bool(False)

    # Is user input set on every keystroke? (Overrides the default) ('simple'
    # style only):
    auto_set = False

    # Is user input set when the Enter key is pressed? (Overrides the default)
    # ('simple' style only):
    enter_set = True

    # The number of history entries to maintain:
    # FIXME: add support
    entries = Int(10)

    # The root path of the file tree view ('custom' style only, not supported
    # under wx). If not specified, the filesystem root is used.
    root_path = File

    # Optional extend trait name of the trait containing the root path.
    root_path_name = Str

    # Optional extended trait name used to notify the editor when the file
    # system view should be reloaded ('custom' style only):
    reload_name = Str

    # Optional extended trait name used to notify when the user double-clicks
    # an entry in the file tree view. The associated path is assigned it:
    dclick_name = Str

    # The style of file dialog to use when the 'Browse...' button is clicked
    # Should be one of 'open' or 'save'
    dialog_style = Str('open')

    #-------------------------------------------------------------------------
    #  Traits view definition:
    #-------------------------------------------------------------------------

    traits_view = View([['<options>',
                         'truncate_ext{Automatically truncate file extension?}',
                         '|options:[Options]>'],
                        ['filter', '|[Wildcard filters]<>']])

    extras = Group()

# Define the FileEditor class.
FileEditor = ToolkitEditorFactory

## EOF ########################################################################



""" Tests for the action manager builder. """

try:
    from pyface.api import GUI  # noqa
except ImportError:
    from nose.plugins.skip import SkipTest
    raise SkipTest('Skipping due to missing GUI toolkit')

# Enthought library imports.
from envisage.ui.action.api import Action, ActionSet, Group, Menu
from traits.testing.unittest_tools import unittest

# Local imports.
from .dummy_action_manager_builder import DummyActionManagerBuilder


class ActionManagerBuilderTestCase(unittest.TestCase):
    """ Tests for the action manager builder. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_action_with_nonexistent_group(self):
        """ action with non-existent group """

        action_sets = [
            ActionSet(
                actions = [
                    Action(
                        class_name = 'Exit',
                        path       = 'MenuBar/File',
                        group      = 'Bogus'
                    ),
                ]
            ),

        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        self.failUnlessRaises(
            ValueError, builder.create_menu_bar_manager, 'MenuBar'
        )

        return

    def test_action_with_nonexistent_sibling(self):
        """ action with non-existent sibling """

        action_sets = [
            ActionSet(
                actions = [
                    Action(
                        class_name = 'Exit',
                        path       = 'MenuBar/File',
                        before     = 'NonExistentAction'
                    ),
                ]
            ),

        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        self.failUnlessRaises(
            ValueError, builder.create_menu_bar_manager, 'MenuBar'
        )

        return

    def test_group_with_nonexistent_sibling(self):
        """ group with non-existent sibling """

        action_sets = [
            ActionSet(
                groups = [
                    Group(id='FileMenuGroup', path='MenuBar', before='Bogus')
                ]
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        self.failUnlessRaises(
            ValueError, builder.create_menu_bar_manager, 'MenuBar'
        )

        return

    def test_menu_with_nonexistent_sibling(self):
        """ menu with non-existent sibling """

        action_sets = [
            ActionSet(
                menus = [
                    Menu(name='&File', path='MenuBar', before='Bogus')
                ]
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        self.failUnlessRaises(
            ValueError, builder.create_menu_bar_manager, 'MenuBar'
        )

        return

    def test_action_with_path_component_that_is_not_a_menu(self):
        """ action with path component that is not a menu """

        action_sets = [
            ActionSet(
                actions = [
                    Action(
                        class_name = 'Exit',
                        path       = 'MenuBar/File'
                    ),

                    Action(
                        class_name = 'Broken',
                        path       = 'MenuBar/File/Exit',
                    ),
                ]
            ),

        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        self.failUnlessRaises(
            ValueError, builder.create_menu_bar_manager, 'MenuBar'
        )

        return

    def test_single_top_level_menu_with_no_group(self):
        """ single top level menu with no group """

        action_sets = [
            ActionSet(
                menus = [
                    Menu(name='&File', path='MenuBar'),
                ]
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_bar_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure that the 'File' menu was added to the 'additions' group
        # of the menubar.
        self.assertEqual(1, len(menu_bar_manager.groups))

        group = menu_bar_manager.find_group('additions')
        ids = [item.id for item in group.items]
        self.assertEqual(['File'], ids)

        return

    def test_single_top_level_group(self):
        """ single top level group """

        action_sets = [
            ActionSet(
                groups = [
                    Group(id='FileMenuGroup', path='MenuBar'),
                ]
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_bar_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure that the group was added before the 'additions' group.
        self.assertEqual(2, len(menu_bar_manager.groups))

        ids = [group.id for group in menu_bar_manager.groups]
        self.assertEqual(['FileMenuGroup', 'additions'], ids)

        return

    def test_top_level_menus_with_no_groups(self):
        """ top level menus with_no groups """

        action_sets = [
            ActionSet(
                menus = [
                    Menu(name='&File', path='MenuBar'),
                    Menu(name='&Edit', path='MenuBar'),
                    Menu(name='&Tools', path='MenuBar'),
                    Menu(name='&Help', path='MenuBar')
                ],
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_bar_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure that all of the menus were added the the 'additions' group
        # of the menubar (and in the right order!).
        self.assertEqual(1, len(menu_bar_manager.groups))

        group = menu_bar_manager.find_group('additions')
        ids = [item.id for item in group.items]
        self.assertEqual(['File', 'Edit', 'Tools', 'Help'], ids)

        return

    def test_top_level_menus_no_groups_before_and_after(self):
        """ top level menus no groups, before and after """

        action_sets = [
            ActionSet(
                menus = [
                    Menu(name='&Edit', path='MenuBar', after='File'),
                ],
            ),

            ActionSet(
                menus = [
                    Menu(name='&File', path='MenuBar'),
                ],
            ),

            ActionSet(
                menus = [
                    Menu(name='&Help', path='MenuBar')
                ],
            ),

            ActionSet(
                menus = [
                    Menu(name='&Tools', path='MenuBar', before='Help'),
                ],
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure that all of the menus were added the the 'additions' group
        # of the menubar.
        self.assertEqual(1, len(menu_manager.groups))

        additions = menu_manager.find_group('additions')
        ids = [item.id for item in additions.items]
        self.assertEqual(['File', 'Edit', 'Tools', 'Help'], ids)

        return

    def test_top_level_menu_non_existent_group(self):
        """ top level menu non-existent group """

        action_sets = [
            ActionSet(
                menus = [
                    Menu(name='&File', path='MenuBar', group='FileMenuGroup'),
                ],
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        self.failUnlessRaises(
            ValueError, builder.create_menu_bar_manager, 'MenuBar'
        )

        return

    def test_top_level_menu_group(self):
        """ top level menu group """

        action_sets = [
            ActionSet(
                groups = [
                    Group(id='FileMenuGroup', path='MenuBar')
                ],

                menus = [
                    Menu(name='&File', path='MenuBar', group='FileMenuGroup'),
                ],
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure that the 'File' menu was added to the 'FileMenuGroup'
        # group of the menubar.
        self.assertEqual(2, len(menu_manager.groups))

        ids = [group.id for group in menu_manager.groups]
        self.assertEqual(['FileMenuGroup', 'additions'], ids)

        group = menu_manager.find_group('FileMenuGroup')
        self.assertEqual('File', group.items[0].id)

        return

    def test_sub_menus_no_groups(self):
        """ sub-menus no groups """

        # We split the contributions into different action sets just because
        # that is how it might end up in an actual application... not because
        # you *have* to split them up this way!
        action_sets = [
            ActionSet(
                menus = [
                    Menu(name='&File', path='MenuBar'),
                    Menu(name='&Edit', path='MenuBar'),
                    Menu(name='&Tools', path='MenuBar'),
                    Menu(name='&Help', path='MenuBar')
                ],
            ),

            ActionSet(
                menus = [
                    Menu(name='&New', path='MenuBar/File'),
                ],
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure the 'New' sub-menu got added to the 'additions' group
        # of the 'File' menu.
        menu = menu_manager.find_item('File')
        additions = menu.find_group('additions')

        self.assertEqual('New', additions.items[0].id)

        return

    def test_actions_no_groups(self):
        """ actions no groups """

        # We split the contributions into different action sets just because
        # that is how it might end up in an actual application... not because
        # you *have* to split them up this way!
        action_sets = [
            ActionSet(
                menus = [
                    Menu(name='&File', path='MenuBar'),
                    Menu(name='&Edit', path='MenuBar'),
                    Menu(name='&Tools', path='MenuBar'),
                    Menu(name='&Help', path='MenuBar')
                ]
            ),

            ActionSet(
                actions = [
                    Action(class_name='Exit', path='MenuBar/File'),
                    Action(class_name='About', path='MenuBar/Help')
                ]
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure the 'ExitAction' action got added to the 'additions' group
        # of the 'File' menu.
        menu = menu_manager.find_item('File')
        additions = menu.find_group('additions')

        self.assertEqual('Exit', additions.items[0].id)

        # Make sure the 'AboutAction' action got added to the 'additions' group
        # of the 'File' menu.
        menu = menu_manager.find_item('Help')
        additions = menu.find_group('additions')

        self.assertEqual('About', additions.items[0].id)

        return

    def test_actions_make_submenus(self):
        """ actions make submenus """

        action_sets = [
            ActionSet(
                actions = [
                    Action(class_name='Folder', path='MenuBar/File/New'),
                    Action(class_name='File', path='MenuBar/File/New')
                ]
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure the 'File' menu got added to the 'additions' group of the
        # menubar.
        self.assertEqual(1, len(menu_manager.groups))

        additions = menu_manager.find_group('additions')
        self.assertEqual('File', additions.items[0].id)

        # Make sure the 'New' sub-menu got added to the 'additions' group
        # of the 'File' menu.
        menu = menu_manager.find_item('File')
        additions = menu.find_group('additions')

        self.assertEqual('New', additions.items[0].id)

        # Make sure the new 'Folder' and 'File' actions got added to the
        # 'additions' group of the 'New' menu.
        menu = menu_manager.find_item('File/New')
        additions = menu.find_group('additions')

        self.assertEqual('Folder', additions.items[0].id)
        self.assertEqual('File', additions.items[1].id)

        return

    def test_actions_make_submenus_before_and_after(self):
        """ actions make submenus before and after """

        action_sets = [
            ActionSet(
                actions = [
                    Action(
                        class_name = 'File',
                        path       = 'MenuBar/File/New',
                        after      = 'Folder'
                    ),

                    Action(
                        class_name = 'Project',
                        path       = 'MenuBar/File/New',
                        before     = 'Folder'
                    ),

                    Action(
                        class_name = 'Folder',
                        path       = 'MenuBar/File/New',
                    ),
                ]
            )
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure the 'File' menu got added to the 'additions' group of the
        # menubar.
        self.assertEqual(1, len(menu_manager.groups))

        additions = menu_manager.find_group('additions')
        self.assertEqual('File', additions.items[0].id)

        # Make sure the 'New' sub-menu got added to the 'additions' group
        # of the 'File' menu.
        menu = menu_manager.find_item('File')
        additions = menu.find_group('additions')

        self.assertEqual('New', additions.items[0].id)

        # Make sure the new 'Folder' and 'File' actions got added to the
        # 'additions' group of the 'New' menu.
        menu = menu_manager.find_item('File/New')
        additions = menu.find_group('additions')

        ids = [item.id for item in additions.items]
        self.assertEqual(['Project', 'Folder', 'File'], ids)

        return

    def test_explicit_groups(self):
        """ explicit groups """

        action_sets = [
            ActionSet(
                menus = [
                    Menu(name='&File', path='MenuBar'),
                    Menu(name='&Edit', path='MenuBar'),
                    Menu(name='&Tools', path='MenuBar'),
                    Menu(name='&Help', path='MenuBar')
                ],
            ),

            ActionSet(
                menus = [
                    Menu(name='&New', path='MenuBar/File', group='NewGroup'),
                ],
            ),

            ActionSet(
                actions = [
                    Action(
                        class_name = 'Exit',
                        path       = 'MenuBar/File',
                        group      = 'ExitGroup'
                    ),
                ]
            ),

            ActionSet(
                groups = [
                    Group(
                        id   = 'ExitGroup',
                        path = 'MenuBar/File'
                    ),

                    Group(
                        id    = 'SaveGroup',
                        path  = 'MenuBar/File',
                        after = 'NewGroup'
                    ),

                    Group(
                        id     = 'NewGroup',
                        path   = 'MenuBar/File',
                        before = 'ExitGroup'
                    ),

                ]
            ),

        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure that all of the menus were added the the 'additions' group
        # of the menubar.
        self.assertEqual(1, len(menu_manager.groups))

        additions = menu_manager.find_group('additions')
        ids = [item.id for item in additions.items]
        self.assertEqual(['File', 'Edit', 'Tools', 'Help'], ids)

        # Make sure the 'File' menu has got 3 groups, 'NewGroup', 'ExitGroup'
        # and 'additions' (and in that order!).
        menu = menu_manager.find_item('File')
        self.assertEqual(4, len(menu.groups))

        ids = [group.id for group in menu.groups]
        self.assertEqual(
            ['NewGroup', 'SaveGroup', 'ExitGroup', 'additions'], ids
        )

        # Make sure the 'New' sub-menu got added to the 'NewGroup' group
        # of the 'File' menu.
        menu = menu_manager.find_item('File')
        group = menu.find_group('NewGroup')

        self.assertEqual('New', group.items[0].id)

        # Make sure the 'Exit' action got added to the 'ExitGroup' group
        # of the 'File' menu.
        menu = menu_manager.find_item('File')
        group = menu.find_group('ExitGroup')

        self.assertEqual('Exit', group.items[0].id)

        return

    def test_actions_and_menus_in_groups(self):
        """ actions and menus in groups """

        action_sets = [
            ActionSet(
                menus = [
                    Menu(
                        name   = '&File',
                        path   = 'MenuBar',
                        groups = ['NewGroup', 'ExitGroup']
                    ),

                    Menu(name='&Edit', path='MenuBar'),
                    Menu(name='&Tools', path='MenuBar'),
                    Menu(name='&Help', path='MenuBar')
                ],
            ),

            ActionSet(
                menus = [
                    Menu(name='&New', path='MenuBar/File', group='NewGroup'),
                ],
            ),

            ActionSet(
                actions = [
                    Action(
                        class_name = 'Exit',
                        path       = 'MenuBar/File',
                        group      = 'ExitGroup'
                    ),
                ]
            ),

        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure that all of the menus were added the the 'additions' group
        # of the menubar.
        self.assertEqual(1, len(menu_manager.groups))

        additions = menu_manager.find_group('additions')
        ids = [item.id for item in additions.items]
        self.assertEqual(['File', 'Edit', 'Tools', 'Help'], ids)

        # Make sure the 'File' menu has got 3 groups, 'NewGroup', 'ExitGroup'
        # and 'additions' (and in that order!).
        menu = menu_manager.find_item('File')
        self.assertEqual(3, len(menu.groups))

        ids = [group.id for group in menu.groups]
        self.assertEqual(['NewGroup', 'ExitGroup', 'additions'], ids)

        # Make sure the 'New' sub-menu got added to the 'NewGroup' group
        # of the 'File' menu.
        menu = menu_manager.find_item('File')
        group = menu.find_group('NewGroup')

        self.assertEqual('New', group.items[0].id)

        # Make sure the 'Exit' action got added to the 'ExitGroup' group
        # of the 'File' menu.
        menu = menu_manager.find_item('File')
        group = menu.find_group('ExitGroup')

        self.assertEqual('Exit', group.items[0].id)

        return

    def test_duplicate_menu(self):
        """ duplicate menu """

        action_sets = [
            ActionSet(
                menus = [
                    Menu(
                        name   = '&File',
                        path   = 'MenuBar',
                        groups = ['NewGroup', 'ExitGroup']
                    ),
                ],
            ),

            ActionSet(
                menus = [
                    Menu(
                        name   = '&File',
                        path   = 'MenuBar',
                        groups = ['ExtraGroup']
                    ),
                ],
            ),
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure that all of the menus were added the the 'additions' group
        # of the menubar.
        self.assertEqual(1, len(menu_manager.groups))

        # Make sure we only get *one* 'File' menu.
        additions = menu_manager.find_group('additions')
        ids = [item.id for item in additions.items]
        self.assertEqual(['File'], ids)

        # Make sure the 'File' menu has got 4 groups, 'NewGroup', 'ExitGroup',
        # 'ExtraGroup' and 'additions' (and in that order!).
        menu = menu_manager.find_item('File')
        self.assertEqual(4, len(menu.groups))

        ids = [group.id for group in menu.groups]
        self.assertEqual(
            ['NewGroup', 'ExitGroup', 'ExtraGroup', 'additions'], ids
        )

        return

    def test_duplicate_group(self):
        """ duplicate group """

        action_sets = [
            ActionSet(
                menus = [
                    Menu(
                        name   = '&File',
                        path   = 'MenuBar',
                        groups = ['NewGroup', 'ExitGroup']
                    ),
                ],
            ),

            ActionSet(
                menus = [
                    Menu(
                        name   = '&File',
                        path   = 'MenuBar',
                        groups = ['NewGroup']
                    ),
                ],
            ),
        ]

        # Create a builder containing the action set.
        builder = DummyActionManagerBuilder(action_sets=action_sets)

        # Create a menu bar manager for the 'MenuBar'.
        menu_manager = builder.create_menu_bar_manager('MenuBar')

        # Make sure that all of the menus were added the the 'additions' group
        # of the menubar.
        self.assertEqual(1, len(menu_manager.groups))

        # Make sure we only get *one* 'File' menu.
        additions = menu_manager.find_group('additions')
        ids = [item.id for item in additions.items]
        self.assertEqual(['File'], ids)

        # Make sure the 'File' menu has got 3 groups, 'NewGroup', 'ExitGroup'
        # and 'additions' (and in that order!).
        menu = menu_manager.find_item('File')
        self.assertEqual(3, len(menu.groups))

        ids = [group.id for group in menu.groups]
        self.assertEqual(
            ['NewGroup', 'ExitGroup', 'additions'], ids
        )

        return


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



import OpenGL.GL as gl

active_shader = None


class Shader(object):

    """An OpenGL shader object"""
    def __init__(self, shader_type, src=None, name="(unnamed shader)"):
        self.shader_object = gl.glCreateShader(shader_type)
        self.name = name
        self.program = None

        if src:
            self.source(src)
            self.compile()
    
    def __del__(self):
        if self.program:
            self.program.detach(self)
            self.program = None
        gl.glDeleteShader(self.shader_object)
    
    def source(self, source_string):
        gl.glShaderSource(self.shader_object, source_string)
    
    def compile(self):
        gl.glCompileShader(self.shader_object)
        return_val = gl.glGetShaderiv(self.shader_object, gl.GL_COMPILE_STATUS)

        if return_val:
            print "%s compiled successfuly." % (self.name)
        else:
            print "Compile failed on shader %s: " % (self.name)
            print gl.glGetShaderInfoLog(self.shader_object)

    def info_log(self):
        return gl.glGetProgramInfoLog(self.shader_object)

    def print_info_log(self):
        print self.info_log()


class UniformVar(object):
    def __init__(self, set_function, name, *args ):
        self.set_function = set_function
        self.name = name
        self.values = args
    
    def set(self):
        self.set_function( self.name, *self.values )


class Program(object):
    """An OpenGL shader program"""
    def __init__(self, shaders=None):
        self.program_object = gl.glCreateProgram()
        self.shaders = []
        self.uniform_vars = {}

        if shaders:
            for s in shaders:
                self.attach(s)
            self.link()
            self.use()
            self.stop()
    
    def __del__(self):
        gl.glDeleteProgram(self.program_object)
    
    def attach(self, shader):
        self.shaders.append(shader)
        shader.program = self
        gl.glAttachShader(self.program_object, shader.shader_object)
    
    def detach(self, shader):
        self.shaders.remove(shader)
        gl.glDetachShader(self.program_object, shader.shader_object)
        print "Shader detached"
    
    def link(self):
        gl.glLinkProgram(self.program_object)
    
    def use(self):
        global active_shader
        active_shader = self
        gl.glUseProgram( self.program_object )
        self.set_vars()

    def stop(self):
        global active_shader
        gl.glUseProgram( 0 )
        active_shader = None

    def uniformi( self, name, *args ):
        argf = {1: gl.glUniform1i,
                2: gl.glUniform2i,
                3: gl.glUniform3i,
                4: gl.glUniform4i}
        f = argf[len(args)]

        def _set_uniform(name, *args):
            location = gl.glGetUniformLocation(self.program_object, name)
            f(location, *args)

        self.uniform_vars[name] = UniformVar(_set_uniform, name, *args)
        if self == active_shader:
            self.uniform_vars[name].set()
    
    def uniformf(self, name, *args):
        argf = {1: gl.glUniform1f,
                2: gl.glUniform2f,
                3: gl.glUniform3f,
                4: gl.glUniform4f}
        f = argf[len(args)]

        def _set_uniform(name, *args):
            location = gl.glGetUniformLocation(self.program_object, name)
            f(location, *args)

        self.uniform_vars[name] = UniformVar(_set_uniform, name, *args)
        if self == active_shader:
            self.uniform_vars[name].set()
    
    def uniform_matrixf(self, name, transpose, values):
        argf = {4: gl.glUniformMatrix2fv,
                9: gl.glUniformMatrix3fv,
                16: gl.glUniformMatrix4fv}
        f = argf[len(values)]

        def _set_uniform(name, values):
            location = gl.glGetUniformLocation(self.program_object, name)
            f(location, 1, transpose, values)

        self.uniform_vars[name] = UniformVar(_set_uniform, name, values)
        if self == active_shader:
            self.uniform_vars[name].set()
    
    def set_vars(self):
        for name, var in self.uniform_vars.iteritems():
            var.set()
    
    def print_info_log(self):
        print gl.glGetInfoLog(self.program_object)


def make_ps_from_src(name, src):
    return make_shader_from_src(name, src, gl.GL_FRAGMENT_SHADER)


def make_vs_from_src(name, src):
    return make_shader_from_src(name, src, gl.GL_VERTEX_SHADER)


def make_shader_from_src(name, src, shader_type):
    return Shader(shader_type, src=src, name=name)


def make_program_from_src_files(vertex_shader_name, pixel_shader_name):
    with open(vertex_shader_name, "r") as f:
        vs_src = f.tostring()
    with open(pixel_shader_name, "r") as f:
        ps_src = f.tostring()
    return make_program_from_src(vertex_shader_name, pixel_shader_name, vs_src, ps_src)


def make_program_from_src(vs_name, ps_name, vertex_shader_src, pixel_shader_src):
    vs = make_vs_from_src(vs_name, vertex_shader_src)
    ps = make_ps_from_src(ps_name, pixel_shader_src)
    p = Program([vs, ps])
    return p


def disable_shaders():
    global active_shader
    gl.glUseProgram(0)
    active_shader = None



from .errors import (IDConflictError, ListenerNotConnectedError,
                     ProviderNotRegisteredError)
from .i_selection import ISelection, IListSelection
from .i_selection_provider import ISelectionProvider
from .list_selection import ListSelection
from .selection_service import SelectionService



#!/usr/bin/env python

"""
build64.py

genereates a text file with the base64encoded contentes of a file

"""

import sys, base64

infilename = sys.argv[1]


outfilename = infilename + ".b64"

data = file(infilename, 'r').read()

data = base64.b64encode(data)

file(outfilename,'w').write(data)

file('junk.png','w').write(base64.b64decode(data))

print data




import numpy as np
from numpy.testing import assert_allclose
from pandas import DataFrame

from ..table_node import H5TableNode
from .utils import temp_h5_file


NODE = '/table_node'


def test_basics():
    description = [('a', np.float64), ('b', np.float64)]
    with temp_h5_file() as h5:
        h5table = H5TableNode.add_to_h5file(h5, NODE, description)
        h5table.append({'a': [1, 2], 'b': [3, 4]})

        assert_allclose(h5table['a'], [1, 2])
        assert_allclose(h5table['b'], [3, 4])

    dtype_description = np.dtype([('c', 'f4'), ('d', 'f4')])
    with temp_h5_file() as h5:
        h5table = H5TableNode.add_to_h5file(h5, NODE, dtype_description)
        h5table.append({'c': [1.2, 3.4], 'd': [5.6, 7.8]})

        assert_allclose(h5table['c'], [1.2, 3.4])
        assert_allclose(h5table['d'], [5.6, 7.8])

        assert len(repr(h5table)) > 0


def test_getitem():
    description = [('a', np.float64), ('b', np.float64)]
    with temp_h5_file() as h5:
        h5table = H5TableNode.add_to_h5file(h5, NODE, description)
        h5table.append({'a': [1, 2], 'b': [3, 4]})
        assert_allclose(h5table['a'], (1, 2))
        assert_allclose(h5table[['b', 'a']], [(3, 1),
                                              (4, 2)])


def test_keys():
    description = [('hello', 'int'), ('world', 'int'), ('Qux1', 'bool')]
    with temp_h5_file() as h5:
        keys = set(list(zip(*description))[0])
        h5table = H5TableNode.add_to_h5file(h5, NODE, description)
        assert set(h5table.keys()) == keys


def test_to_dataframe():
    description = [('a', np.float64)]
    with temp_h5_file() as h5:
        h5table = H5TableNode.add_to_h5file(h5, NODE, description)
        h5table.append({'a': [1, 2, 3]})
        df = h5table.to_dataframe()
        assert isinstance(df, DataFrame)
        assert_allclose(df['a'], h5table['a'])


if __name__ == '__main__':
    from numpy import testing
    testing.run_module_suite()
